#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Gmsh and related objects
"""

from dessia_common import DessiaObject
import volmdlr
import volmdlr.mesh


class GmshParser(DessiaObject):
    _standalone_in_db = False
    _non_serializable_attributes = []
    _non_eq_attributes = ['name']
    _non_hash_attributes = ['name']
    _generic_eq = True

    def __init__(self,
                 mesh_format: dict,
                 nodes: dict,
                 elements: dict,
                 entities: dict,
                 physical_names=None,
                 partitioned_entities=None,
                 periodic=None,
                 # ghost_elements = None,
                 parametrizations=None,
                 node_data=None,
                 element_data=None,
                 element_node_data=None,
                 # interpolation_scheme = None,
                 name: str = ''):

        self.mesh_format = mesh_format
        self.physical_names = physical_names
        self.entities = entities
        self.nodes = nodes
        self.elements = elements
        self.partitioned_entities = partitioned_entities
        self.periodic = periodic
        # self.ghost_elements = ghost_elements,
        self.parametrizations = parametrizations
        self.node_data = node_data
        self.element_data = element_data
        self.element_node_data = element_node_data
        # self.interpolation_scheme = interpolation_scheme,
        self.name = name

        DessiaObject.__init__(self, name=name)

    @classmethod
    def from_file(cls, file_path: str):
        """
        defines a gmsh object from .msh file
        """

        file_data = GmshParser.read_file(file_path)
        mesh_format = GmshParser.from_file_mesh_format(file_data['MeshFormat'])
        physical_names = GmshParser.from_file_physical_names(file_data['PhysicalNames'])
        entities = GmshParser.from_file_entities(file_data['Entities'])
        nodes = GmshParser.from_file_nodes(file_data['Nodes'])
        elements = GmshParser.from_file_elements(file_data['Elements'])
        partitioned_entities = GmshParser.from_file_partitioned_entities(file_data['PartitionedEntities'])
        periodic = GmshParser.from_file_periodic(file_data['Periodic'])
        # ghost_elements = GmshParser.from_file_ghost_elements(file_data['GhostElements'])
        parametrizations = GmshParser.from_file_parametrizations(file_data['Parametrizations'])
        node_data = GmshParser.from_file_node_data(file_data['NodeData'])
        element_data = GmshParser.from_file_element_data(file_data['ElementData'])
        element_node_data = GmshParser.from_file_element_node_data(file_data['ElementNodeData'])
        # interpolation_scheme = GmshParser.from_file_interpolation_scheme(file_data['InterpolationScheme'])

        return cls(mesh_format=mesh_format,
                   entities=entities,
                   nodes=nodes,
                   elements=elements,
                   physical_names=physical_names,
                   partitioned_entities=partitioned_entities,
                   periodic=periodic,
                   # ghost_elements=ghost_elements,
                   parametrizations=parametrizations,
                   node_data=node_data,
                   element_data=element_data,
                   element_node_data=element_node_data,
                   # interpolation_scheme=interpolation_scheme
                   name='')

    @staticmethod
    def from_file_elements(lines):
        """
        gets elements data from .msh file
        """

        if not lines:
            return {}

        elements = {}
        elements_types, elements_type = [], []
        step = 1
        while True:
            line = lines[step].split()
            num_elements_in_block = int(line[3])
            if num_elements_in_block:
                element_type = line[2]
                try:
                    elements['elements_type_' + element_type]
                except KeyError:
                    elements_types.append(elements_type)
                    elements['elements_type_' + element_type] = []
                    elements_type = []

                step = step + 1
                elements_list = []
                for i in range(step, step + num_elements_in_block):
                    line = lines[i].split()
                    element = [int(index) - 1 for index in line[1::]]
                    # elements['elements_type_'+ element_type].append(element)
                    elements_list.append(element)
                    elements_type.append(element)
                elements['elements_type_' + element_type].append(elements_list)

            step = i + 1  # num_nodes_in_block

            if step == len(lines):
                break

        return elements

    # $ElementData
    #   numStringTags(ASCII int)
    #   stringTag(string) ...
    #   numRealTags(ASCII int)
    #   realTag(ASCII double) ...
    #   numIntegerTags(ASCII int)
    #   integerTag(ASCII int) ...
    #   elementTag(int) value(double) ...
    #   ...
    # $EndElementData

    @staticmethod
    def from_file_element_data(lines):
        """
        gets mesh element_data from .msh file
        """

        if not lines:
            return {}

        element_data = {}

        return element_data

    # $ElementNodeData
    #   numStringTags(ASCII int)
    #   stringTag(string) ...
    #   numRealTags(ASCII int)
    #   realTag(ASCII double) ...
    #   numIntegerTags(ASCII int)
    #   integerTag(ASCII int) ...
    #   elementTag(int) numNodesPerElement(int) value(double) ...
    #   ...
    # $EndElementNodeData

    @staticmethod
    def from_file_element_node_data(lines):
        """
        gets mesh element_node_data from .msh file
        """

        if not lines:
            return {}

        element_node_data = {}

        return element_node_data

    @staticmethod
    def from_file_entities(lines):
        """
        gets entities data from .msh file
        """

        if not lines:
            return {}

        entities = [int(lines[0].split()[0]), int(lines[0].split()[1]),
                    int(lines[0].split()[2]), int(lines[0].split()[3])]

        points_data = []
        for i in range(1, entities[0] + 1):
            line = lines[i].split()
            points = {}
            points['pointTag'] = int(line[0])
            points['X'] = float(line[1])
            points['Y'] = float(line[2])
            points['Z'] = float(line[3])
            points['numPhysicalTags'] = int(line[4])
            if points['numPhysicalTags']:
                points['physicalTag'] = int(line[5])
            points_data.append(points)

        curves_data = []
        for i in range(entities[0] + 1, entities[0] + entities[1] + 1):
            line = lines[i].split()
            curves = {}
            curves['curveTag'] = int(line[0])
            curves['minX'] = float(line[1])
            curves['minY'] = float(line[2])
            curves['minZ'] = float(line[3])
            curves['maxX'] = float(line[4])
            curves['maxY'] = float(line[5])
            curves['maxZ'] = float(line[6])
            curves['numPhysicalTags'] = int(line[7])
            if curves['numPhysicalTags']:
                curves['physicalTag'] = int(line[8])
                curves['numBoundingPoints'] = int(line[9])
                curves['pointTag'] = [float(l) for l in line[10::]]
            else:
                curves['numBoundingPoints'] = int(line[8])
                curves['pointTag'] = [float(l) for l in line[9::]]
            curves_data.append(curves)

        surfaces_data = []
        for i in range(entities[0] + entities[1] + 1, entities[0] + entities[1] + entities[2] + 1):
            line = lines[i].split()
            surfaces = {}
            surfaces['surfaceTag'] = int(line[0])
            surfaces['minX'] = float(line[1])
            surfaces['minY'] = float(line[2])
            surfaces['minZ'] = float(line[3])
            surfaces['maxX'] = float(line[4])
            surfaces['maxY'] = float(line[5])
            surfaces['maxZ'] = float(line[6])
            surfaces['numPhysicalTags'] = int(line[7])
            if surfaces['numPhysicalTags']:
                surfaces['physicalTag'] = int(line[8])
                surfaces['numBoundingCurves'] = int(line[9])
                surfaces['curveTag'] = [float(l) for l in line[10::]]
            else:
                surfaces['numBoundingCurves'] = float(line[8])
                surfaces['curveTag'] = [float(l) for l in line[9::]]
            surfaces_data.append(surfaces)

        volumes_data = []
        for i in range(entities[0] + entities[1] + entities[2] + 1, len(lines)):
            line = lines[i].split()
            volumes = {}
            volumes['volumeTag'] = int(line[0])
            volumes['minX'] = float(line[1])
            volumes['minY'] = float(line[2])
            volumes['minZ'] = float(line[3])
            volumes['maxX'] = float(line[4])
            volumes['maxY'] = float(line[5])
            volumes['maxZ'] = float(line[6])
            volumes['numPhysicalTags'] = int(line[7])
            if volumes['numPhysicalTags']:
                volumes['physicalTag'] = int(line[8])
                volumes['numBoundngSurfaces'] = int(line[9])
                volumes['surfaceTag'] = [float(l) for l in line[10::]]
            else:
                volumes['numBoundngSurfaces'] = float(line[8])
                volumes['surfaceTag'] = [float(l) for l in line[9::]]
            volumes_data.append(volumes)

        return {'points': points_data,
                'curves': curves_data,
                'surfaces': surfaces_data,
                'volumes': volumes_data}

    # $GhostElements
    #   numGhostElements(size_t)
    #   elementTag(size_t) partitionTag(int)
    #     numGhostPartitions(size_t) ghostPartitionTag(int) ...
    #   ...
    # $EndGhostElements

    @staticmethod
    def from_file_ghost_elements(lines):
        """
        gets mesh ghost_elements from .msh file
        """

        if not lines:
            return {}

        ghost_elements = {}

        return ghost_elements

    # $InterpolationScheme
    #   name(string)
    #   numElementTopologies(ASCII int)
    #   elementTopology
    #   numInterpolationMatrices(ASCII int)
    #   numRows(ASCII int) numColumns(ASCII int) value(ASCII double) ...
    # $EndInterpolationScheme

    @staticmethod
    def from_file_interpolation_scheme(lines):
        """
        gets mesh interpolation_scheme from .msh file
        """

        if not lines:
            return {}

        interpolation_scheme = {}

        return interpolation_scheme

    @staticmethod
    def from_file_mesh_format(lines):
        """
        gets mesh format data from .msh file
        """

        mesh_format = {}
        mesh_format['version_number'] = float(lines[0].split()[0])
        if int(lines[0].split()[1]) == 0:
            mesh_format['file_type'] = (int(lines[0].split()[1]), 'ASCII')
        else:
            mesh_format['file_type'] = (int(lines[0].split()[1]), 'Binary')
        mesh_format['data_size'] = int(lines[0].split()[2])

        return mesh_format

    @staticmethod
    def from_file_nodes(lines):
        """
        gets mesh nodes from .msh file
        """

        if not lines:
            return {}

        nodes = {}
        nodes_points = []
        step = 1
        while True:
            line = lines[step].split()
            num_nodes_in_block = int(line[3])
            if num_nodes_in_block:
                entity_dim = line[0]
                # nodes['nodes_dim_'+ entity_dim] = []
                try:
                    nodes['nodes_dim_' + entity_dim]
                except KeyError:
                    nodes['nodes_dim_' + entity_dim] = []

                step = step + num_nodes_in_block + 1
                points = []
                for i in range(step, step + num_nodes_in_block):
                    line = lines[i].split()
                    points.append(volmdlr.mesh.Node3D(float(line[0]),
                                                      float(line[1]),
                                                      float(line[2])))

                    # nodes['nodes_dim_'+ entity_dim].append(
                    #     volmdlr.Point3D(float(line[0]),
                    #                     float(line[1]),
                    #                     float(line[2])))
                    # nodes_points.append(volmdlr.Point3D(float(line[0]),
                    #                                     float(line[1]),
                    #                                     float(line[2])))

                nodes['nodes_dim_' + entity_dim].append(points)
                nodes_points.extend(points)

                step = step + num_nodes_in_block

            else:
                step += 1

            if len(nodes_points) == int(lines[0].split()[1]):
                break

        if GmshParser.check_2d(nodes_points):
            for key, value in nodes.items():
                values = []
                for points in value:
                    values.append(GmshParser.to_2d(points))
            nodes[key] = values
            nodes['all_nodes'] = GmshParser.to_2d(nodes_points)
        else:
            nodes['all_nodes'] = nodes_points

        return nodes

    # $NodeData
    #   numStringTags(ASCII int)
    #   stringTag(string) ...
    #   numRealTags(ASCII int)
    #   realTag(ASCII double) ...
    #   numIntegerTags(ASCII int)
    #   integerTag(ASCII int) ...
    #   nodeTag(int) value(double) ...
    #   ...
    # $EndNodeData

    @staticmethod
    def from_file_node_data(lines):
        """
        gets mesh node_data from .msh file
        """

        if not lines:
            return {}

        node_data = {}

        return node_data

    # $Parametrizations
    #   numCurveParam(size_t) numSurfaceParam(size_t)
    #   curveTag(int) numNodes(size_t)
    #     nodeX(double) nodeY(double) nodeZ(double) nodeU(double)
    #     ...
    #   ...
    #   surfaceTag(int) numNodes(size_t) numTriangles(size_t)
    #     nodeX(double) nodeY(double) nodeZ(double)
    #       nodeU(double) nodeV(double)
    #       curvMaxX(double) curvMaxY(double) curvMaxZ(double)
    #       curvMinX(double) curvMinY(double) curvMinZ(double)
    #     ...
    #     nodeIndex1(int) nodeIndex2(int) nodeIndex3(int)
    #     ...
    #   ...
    # $EndParametrizations

    @staticmethod
    def from_file_parametrizations(lines):
        """
        gets mesh parametrizations from .msh file
        """

        if not lines:
            return {}

        parametrizations = {}

        return parametrizations

    # $PartitionedEntities
    #   numPartitions(size_t)
    #   numGhostEntities(size_t)
    #     ghostEntityTag(int) partition(int)
    #     ...
    #   numPoints(size_t) numCurves(size_t)
    #     numSurfaces(size_t) numVolumes(size_t)
    #   pointTag(int) parentDim(int) parentTag(int)
    #     numPartitions(size_t) partitionTag(int) ...
    #     X(double) Y(double) Z(double)
    #     numPhysicalTags(size_t) physicalTag(int) ...
    #   ...
    #   curveTag(int) parentDim(int) parentTag(int)
    #     numPartitions(size_t) partitionTag(int) ...
    #     minX(double) minY(double) minZ(double)
    #     maxX(double) maxY(double) maxZ(double)
    #     numPhysicalTags(size_t) physicalTag(int) ...
    #     numBoundingPoints(size_t) pointTag(int) ...
    #   ...
    #   surfaceTag(int) parentDim(int) parentTag(int)
    #     numPartitions(size_t) partitionTag(int) ...
    #     minX(double) minY(double) minZ(double)
    #     maxX(double) maxY(double) maxZ(double)
    #     numPhysicalTags(size_t) physicalTag(int) ...
    #     numBoundingCurves(size_t) curveTag(int) ...
    #   ...
    #   volumeTag(int) parentDim(int) parentTag(int)
    #     numPartitions(size_t) partitionTag(int) ...
    #     minX(double) minY(double) minZ(double)
    #     maxX(double) maxY(double) maxZ(double)
    #     numPhysicalTags(size_t) physicalTag(int) ...
    #     numBoundingSurfaces(size_t) surfaceTag(int) ...
    #   ...
    # $EndPartitionedEntities

    @staticmethod
    def from_file_partitioned_entities(lines):
        """
        gets mesh partitioned_entities from .msh file
        """

        if not lines:
            return {}

        partitioned_entities = {}

        return partitioned_entities

    # $Periodic
    #   numPeriodicLinks(size_t)
    #   entityDim(int) entityTag(int) entityTagMaster(int)
    #   numAffine(size_t) value(double) ...
    #   numCorrespondingNodes(size_t)
    #     nodeTag(size_t) nodeTagMaster(size_t)
    #     ...
    #   ...
    # $EndPeriodic

    @staticmethod
    def from_file_periodic(lines):
        """
        gets mesh periodic from .msh file
        """

        if not lines:
            return {}

        periodic = {}

        return periodic

    @staticmethod
    def from_file_physical_names(lines):
        """
        gets mesh physical_names from .msh file
        """

        if not lines:
            return {}

        physical_names = {}
        for i in range(1, int(lines[0].split()[0]) + 1):
            physical_dim = (lines[i].split()[0])
            try:
                physical_names['physical_dim_' + physical_dim]
            except KeyError:
                physical_names['physical_dim_' + physical_dim] = []

            physical_names['physical_dim_' + physical_dim].append({'physicalTag': int(lines[1][0]),
                                                                   'Name': lines[i][5:-1]})

        return physical_names

    @staticmethod
    def read_file(file_path: str):
        """
        gets lines from a .msh file
        """

        data = {'MeshFormat': [],
                'PhysicalNames': [],
                'Entities': [],
                'PartitionedEntities': [],
                'Nodes': [],
                'Elements': [],
                'Periodic': [],
                'GhostElements': [],
                'Parametrizations': [],
                'NodeData': [],
                'ElementData': [],
                'ElementNodeData': [],
                'InterpolationScheme': []}

        f = open(file_path, "r", encoding="utf-8")
        # lines = []
        while True:
            line = f.readline().strip()
            if not line:
                break
            if line[0] == '$':
                data_type = line[1::]
                line = f.readline().strip()
                while line[0:4] != '$End':
                    # lines.append(line)
                    data[data_type].append(line)
                    line = f.readline().strip()
                # data[data_type] = lines
                # lines = []

        return data

    def define_tetrahedron_element_mesh(self):
        """
        defines a volmdlr mesh with TetrahedronElement from a .msh file
        """

        # nodes = self.nodes[0]
        points = self.nodes['all_nodes']
        # elements = self.elements[0]

        tetrahedron_elements = self.elements['elements_type_4']
        element_groups = []
        for tetrahedrons in tetrahedron_elements:
            tetrahedrons_mesh = []
            for tetrahedron in tetrahedrons:
                tetrahedrons_mesh.append(volmdlr.mesh.TetrahedralElement([points[tetrahedron[0]],
                                                                          points[tetrahedron[1]],
                                                                          points[tetrahedron[2]],
                                                                          points[tetrahedron[3]]]))

            element_groups.append(volmdlr.mesh.ElementsGroup(tetrahedrons_mesh, name=''))

        mesh = volmdlr.mesh.Mesh(element_groups)
        # mesh.nodes = points #gmsh points are duplicated > not needed
        # mesh.node_to_index = {mesh.nodes[i]: i for i in range(len(mesh.nodes))}

        return mesh

    def define_triangular_element_mesh(self):
        """
        defines a volmdlr mesh with TriangularElement from a .msh file
        """

        # nodes = self.nodes[0]
        points = self.nodes['all_nodes']
        # elements = self.elements[0]

        triangles_elements = self.elements['elements_type_2']
        element_groups = []
        for triangles in triangles_elements:
            triangles_mesh = []
            for triangle in triangles:
                if points[0].__class__.__name__[-2::] == '3D':
                    triangles_mesh.append(volmdlr.mesh.TriangularElement3D([points[triangle[0]],
                                                                            points[triangle[1]],
                                                                            points[triangle[2]]]))
                else:
                    triangles_mesh.append(volmdlr.mesh.TriangularElement2D([points[triangle[0]],
                                                                            points[triangle[1]],
                                                                            points[triangle[2]]]))
            element_groups.append(volmdlr.mesh.ElementsGroup(triangles_mesh, name=''))

        mesh = volmdlr.mesh.Mesh(element_groups)
        # mesh.nodes = points #gmsh points are duplicated > not needed
        # mesh.node_to_index = {mesh.nodes[i]: i for i in range(len(mesh.nodes))}

        return mesh

    @staticmethod
    def check_2d(list_nodes):
        for node in list_nodes:
            if node[2] != 0:
                return False
        return True

    @staticmethod
    def to_2d(list_nodes):
        return [volmdlr.mesh.Node2D(node[0], node[1]) for
                node in list_nodes]



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""

"""
# from binaryornot.check import is_binary
import io
import struct
import warnings
# from typing import BinaryIO
from typing import List

from binaryornot.check import is_binary
# import kaitaistruct
from kaitaistruct import KaitaiStream

import dessia_common as dc
import volmdlr as vm
import volmdlr.core as vmc
import volmdlr.faces as vmf


# from kaitaistruct import KaitaiStream


class Stl(dc.DessiaObject):
    """STL files are used to represent simple 3D models, defined using
    triangular 3D faces.

    Initially it was introduced as native format for 3D Systems
    Stereolithography CAD system, but due to its extreme simplicity, it
    was adopted by a wide range of 3D modelling, CAD, rapid prototyping
    and 3D printing applications as the simplest 3D model exchange
    format.

    STL is extremely bare-bones format: there are no complex headers, no
    texture / color support, no units specifications, no distinct vertex
    arrays. Whole model is specified as a collection of triangular
    faces.

    There are two versions of the format (text and binary), this spec
    describes binary version.
    """

    _dessia_methods = ['from_text_stream', 'from_text_stream', 'to_closed_shell', 'to_open_shell']

    def __init__(self, triangles: List[vmf.Triangle3D], name: str = ''):
        self.triangles = triangles
        dc.DessiaObject.__init__(self, name=name)

        self.normals = None

    @classmethod
    def points_from_file(cls, filename: str, distance_multiplier=0.001):
        if is_binary(filename):
            with open(filename, 'rb') as file:
                stream = KaitaiStream(file)
                _ = stream.read_bytes(80).decode('utf8')
                num_triangles = stream.read_u4le()

                all_points = []
                for i in range(num_triangles):
                    if i % 5000 == 0:
                        print('reading stl',
                              round(i / num_triangles * 100, 2), '%')
                    # First is normal, unused
                    _ = vm.Vector3D(stream.read_f4le(),
                                    stream.read_f4le(),
                                    stream.read_f4le())
                    p1 = vm.Point3D(distance_multiplier * stream.read_f4le(),
                                    distance_multiplier * stream.read_f4le(),
                                    distance_multiplier * stream.read_f4le())
                    p2 = vm.Point3D(distance_multiplier * stream.read_f4le(),
                                    distance_multiplier * stream.read_f4le(),
                                    distance_multiplier * stream.read_f4le())
                    p3 = vm.Point3D(distance_multiplier * stream.read_f4le(),
                                    distance_multiplier * stream.read_f4le(),
                                    distance_multiplier * stream.read_f4le())
                    all_points.extend([p1, p2, p3])

                    stream.read_u2le()
        return all_points

    @classmethod
    def from_binary_stream(cls, stream: io.BytesIO, distance_multiplier: float = 0.001):
        stream.seek(0)

        stream = KaitaiStream(stream)
        name_slice = stream.read_bytes(80)
        try:
            name = name_slice.decode('utf-8')
        except UnicodeDecodeError:
            name = name_slice.decode('latin-1')

        num_triangles = stream.read_u4le()
        # print(num_triangles)

        triangles = [None] * num_triangles
        invalid_triangles = []
        for i in range(num_triangles):
            if i % 5000 == 0:
                print('reading stl',
                      round(i / num_triangles * 100, 2), '%')
            _ = vm.Vector3D(stream.read_f4le(),
                            stream.read_f4le(),
                            stream.read_f4le())
            p1 = vm.Point3D(distance_multiplier * stream.read_f4le(),
                            distance_multiplier * stream.read_f4le(),
                            distance_multiplier * stream.read_f4le())
            p2 = vm.Point3D(distance_multiplier * stream.read_f4le(),
                            distance_multiplier * stream.read_f4le(),
                            distance_multiplier * stream.read_f4le())
            p3 = vm.Point3D(distance_multiplier * stream.read_f4le(),
                            distance_multiplier * stream.read_f4le(),
                            distance_multiplier * stream.read_f4le())
            try:
                triangles[i] = vmf.Triangle3D(p1, p2, p3)
            except ZeroDivisionError:
                invalid_triangles.append(i)

            stream.read_u2le()
            # print(abr)
        if invalid_triangles:
            # print('invalid_triangles number: ', len(invalid_triangles))
            for i in invalid_triangles[::-1]:
                del triangles[i]
        return cls(triangles, name=name)

    @classmethod
    def from_text_stream(cls, stream: io.StringIO,
                         distance_multiplier: float = 0.001):
        stream.seek(0)

        header = stream.readline()
        name = header[6:]
        triangles = []
        points = []
        for line in stream.readlines():
            if 'vertex' in line:
                line = line.replace('vertex', '')
                line = line.lstrip(' ')
                x, y, z = line.split(' ')
                points.append(vm.Point3D(distance_multiplier * float(x),
                                         distance_multiplier * float(y),
                                         distance_multiplier * float(z)))
            if 'endfacet' in line:
                try:
                    triangles.append(vmf.Triangle3D(points[0],
                                                    points[1],
                                                    points[2]))
                except ZeroDivisionError:
                    pass
                points = []
        return cls(triangles, name=name)

    @classmethod
    def from_file(cls, filename: str = None,
                  distance_multiplier: float = 0.001):
        warnings.warn("Use load_from_file instead of from_file",
                      DeprecationWarning)
        return cls.load_from_file(filename, distance_multiplier)

    @classmethod
    def load_from_file(cls, filepath: str, distance_multiplier: float = 0.001):
        if is_binary(filepath):
            with open(filepath, 'rb') as file:
                return cls.from_binary_stream(
                    file, distance_multiplier=distance_multiplier)

        with open(filepath, 'r', errors='ignore') as file:
            return cls.from_text_stream(
                file, distance_multiplier=distance_multiplier)

    # Commented because seemed invalid
    # @classmethod
    # def from_file_points(cls, filename:str, distance_multiplier=0.001):
    #     all_points = []
    #     if is_binary(filename):
    #         with open(filename, 'rb') as file:
    #             stream = KaitaiStream(file)
    #             name = stream.read_bytes(80).decode('utf8')
    #             # print(name)
    #             num_triangles = stream.read_u4le()
    #             # print(num_triangles)
    #             all_points = []
    #             for i in range(num_triangles):
    #                 if i % 5000 == 0:
    #                     print('reading stl', round(i/num_triangles*100, 2),
    #                           '%')
    #                 _ = vm.Vector3D(stream.read_f4le(),
    #                                      stream.read_f4le(),
    #                                      stream.read_f4le())
    #                 # print(n)
    #                 p1 = vm.Point3D(distance_multiplier*stream.read_f4le(),
    #                                 distance_multiplier*stream.read_f4le(),
    #                                 distance_multiplier*stream.read_f4le())
    #                 p2 = vm.Point3D(distance_multiplier*stream.read_f4le(),
    #                                 distance_multiplier*stream.read_f4le(),
    #                                 distance_multiplier*stream.read_f4le())
    #                 p3 = vm.Point3D(distance_multiplier*stream.read_f4le(),
    #                                 distance_multiplier*stream.read_f4le(),
    #                                 distance_multiplier*stream.read_f4le())
    #                 # print(p1, p2, p3)
    #                 all_points.extend([p1, p2, p3])
    #                 stream.read_u2le()
    #     return all_points

    def save_to_binary_file(self, filepath, distance_multiplier=1000):
        if not filepath.endswith('.stl'):
            filepath += '.stl'
            print('Adding .stl extension: ', filepath)

        with open(filepath, 'wb') as file:
            self.to_stream(file, distance_multiplier=distance_multiplier)

    def save_to_stream(self, stream, distance_multiplier=1000):
        stream.seek(0)

        BINARY_HEADER = "80sI"
        BINARY_FACET = "12fH"

        # counter = 0
        stream.write(struct.pack(BINARY_HEADER, self.name.encode('utf8'),
                                 len(self.triangles)))
        # counter += 1
        for triangle in self.triangles:
            data = [
                0., 0., 0.,
                distance_multiplier * triangle.point1.x,
                distance_multiplier * triangle.point1.y,
                distance_multiplier * triangle.point1.z,
                distance_multiplier * triangle.point2.x,
                distance_multiplier * triangle.point2.y,
                distance_multiplier * triangle.point2.z,
                distance_multiplier * triangle.point3.x,
                distance_multiplier * triangle.point3.y,
                distance_multiplier * triangle.point3.z,
                0]
            stream.write(struct.pack(BINARY_FACET, *data))

    def to_closed_shell(self):
        return vmf.ClosedShell3D(self.triangles, name=self.name)

    def to_open_shell(self):
        return vmf.OpenShell3D(self.triangles, name=self.name)

    def to_volume_model(self):
        closed_shell = self.to_closed_shell()
        return vmc.VolumeModel([closed_shell], name=self.name)

    def extract_points(self):

        points1 = [t.point1 for t in self.triangles]
        points2 = [t.point2 for t in self.triangles]
        points3 = [t.point3 for t in self.triangles]

        valid_points = vm.Vector3D.remove_duplicate(points1 + points2 + points3)
        return valid_points

    # TODO: decide which algorithm to be used (no _BIS)
    def extract_points_BIS(self, min_distance: float = 0.001):
        points = []
        for i, t in enumerate(self.triangles):
            distance12 = t.point1.point_distance(t.point2)
            distance13 = t.point1.point_distance(t.point3)
            distance23 = t.point2.point_distance(t.point3)
            if distance12 > min_distance:
                n_div = int(distance12 / min_distance)
                for n in range(n_div):
                    new_point = t.point1 + (t.point2 - t.point1) * n / n_div
                    points.append(new_point)
            if distance13 > min_distance:
                n_div = int(distance13 / min_distance)
                for n in range(n_div):
                    new_point = t.point1 + (t.point3 - t.point1) * (n + 1) / n_div
                    points.append(new_point)
            if distance23 > min_distance:
                n_div = int(distance23 / min_distance)
                for n in range(n_div):
                    new_point = t.point2 + (t.point3 - t.point2) * n / n_div
                    points.append(new_point)

        valid_points = vm.Vector3D.remove_duplicate(points)
        return valid_points

    @classmethod
    def from_display_mesh(cls, mesh):
        triangles = []
        for i1, i2, i3 in mesh.triangles:
            triangles.append(vmf.Triangle3D(mesh.points[i1],
                                            mesh.points[i2],
                                            mesh.points[i3]))
        return cls(triangles)

    def get_normals(self):
        """
        Returns
        -------
        points_normals : dictionary
            returns a diction
        """
        points_normals = {}
        normals = []
        for triangle in self.triangles:
            normal = triangle.normal()
            for point in triangle.points:
                if point in list(points_normals.keys()):
                    points_normals[point].append(normal)
                else:
                    points_normals[point] = [normal]
        for key, value in points_normals.items():
            point_normal = vm.O3D
            for point in value:
                point_normal += point
            points_normals[key] = point_normal
            try:
                point_normal.normalize()
            except ZeroDivisionError:
                point_normal = value[0]
                points_normals[key] = point_normal
                point_normal.normalize()
            normals.append(point_normal)
        self.normals = normals

        return points_normals

    def clean_flat_triangles(self) -> 'Stl':
        invalid_triangles = []
        for it, triangles in enumerate(self.triangles):
            if triangles.area() < 1e-12:
                invalid_triangles.append(it)
                print(it, triangles.area())

        triangles = self.triangles[:]
        for invalid_triangle_index in invalid_triangles[::-1]:
            triangles.pop(invalid_triangle_index)
        return Stl(triangles)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""

"""
from typing import List, Tuple
import math
import dessia_common as dc
import volmdlr.edges
# import volmdlr.faces as vmf


class Node2D(volmdlr.Point2D):
    def __hash__(self):
        return int(1e6 * (self.x + self.y))

    def __eq__(self, other_node: 'Node2D'):
        if other_node.__class__.__name__ not in ['Vector2D', 'Point2D',
                                                 'Node2D']:
            return False
        return math.isclose(self.x, other_node.x, abs_tol=1e-06) \
            and math.isclose(self.y, other_node.y, abs_tol=1e-06)

    @classmethod
    def from_point(cls, point2d):
        return cls(point2d.x, point2d.y)


class Node3D(volmdlr.Point3D):
    def __hash__(self):
        return int(1e6 * (self.x + self.y + self.z))

    def __eq__(self, other_node: 'Node3D'):
        if other_node.__class__.__name__ not in ['Vector3D', 'Point3D',
                                                 'Node3D']:
            return False
        return math.isclose(self.x, other_node.x, abs_tol=1e-06) \
            and math.isclose(self.y, other_node.y, abs_tol=1e-06) \
            and math.isclose(self.z, other_node.z, abs_tol=1e-06)

    @classmethod
    def from_point(cls, point3d):
        return cls(point3d.x, point3d.y, point3d.z)


class DisplayMesh(dc.DessiaObject):
    def __init__(self, points, triangles, name=''):

        self.points = points
        self.triangles = triangles
        dc.DessiaObject.__init__(self, name=name)
        self._utd_point_index = False

    def check(self):
        npoints = len(self.points)
        for triangle in self.triangles:
            if max(triangle) >= npoints:
                return False
        return True

    @property
    def point_index(self):
        if not self._utd_point_index:
            self._point_index = {p: i for i, p in enumerate(self.points)}
            self._utd_point_index = True
        return self._point_index

    @classmethod
    def merge_meshes(cls, meshes: List['DisplayMesh']):
        """
        Merge several meshes into one
        """
        # Collect points
        ip = 0
        point_index = {}
        points = []
        if len(meshes) == 1:
            return cls(meshes[0].points, meshes[0].triangles)
        for mesh in meshes:
            for point in mesh.points:
                if point not in point_index:
                    point_index[point] = ip
                    ip += 1
                    points.append(point)

        triangles = []
        for mesh in meshes:
            for i1, i2, i3 in mesh.triangles:
                p1 = mesh.points[i1]
                p2 = mesh.points[i2]
                p3 = mesh.points[i3]
                triangles.append((point_index[p1],
                                  point_index[p2],
                                  point_index[p3]))
        return cls(points, triangles)

    def merge_mesh(self, other_mesh):
        # new_points = self.points[:]
        # new_point_index = self.point_index.copy()
        ip = len(self.points)
        # point_index
        # t1 = time.time()
        for point in other_mesh.points:
            if not point in self.point_index:
                self.point_index[point] = ip
                ip += 1
                self.points.append(point)

        # new_triangles = self.triangles[:]
        # t2 = time.time()
        for i1, i2, i3 in other_mesh.triangles:
            p1 = other_mesh.points[i1]
            p2 = other_mesh.points[i2]
            p3 = other_mesh.points[i3]
            self.triangles.append((self._point_index[p1],
                                   self._point_index[p2],
                                   self._point_index[p3]))

        # t3 = time.time()
        # print('t', t2-t1, t3-t2)
        # self._point_index = new_point_index

    def __add__(self, other_mesh):
        new_points = self.points[:]
        new_point_index = self.point_index.copy()
        ip = len(new_points)
        for point in other_mesh.points:
            if not point in new_point_index:
                new_point_index[point] = ip
                ip += 1
                new_points.append(point)

        new_triangles = self.triangles[:]
        for i1, i2, i3 in other_mesh.triangles:
            p1 = other_mesh.points[i1]
            p2 = other_mesh.points[i2]
            p3 = other_mesh.points[i3]
            new_triangles.append((new_point_index[p1],
                                  new_point_index[p2],
                                  new_point_index[p3]))

        return self.__class__(new_points, new_triangles)

    def plot(self, ax=None, numbering=False):
        for ip, point in enumerate(self.points):
            ax = point.plot(ax=ax)
            if numbering:
                ax.text(*point, 'node {}'.format(ip + 1),
                        ha='center', va='center')

        for i1, i2, i3 in self.triangles:
            self._linesegment_class(self.points[i1], self.points[i2]).plot(
                ax=ax)
            self._linesegment_class(self.points[i2], self.points[i3]).plot(
                ax=ax)
            self._linesegment_class(self.points[i1], self.points[i3]).plot(
                ax=ax)

        return ax


class DisplayMesh2D(DisplayMesh):
    _linesegment_class = volmdlr.edges.LineSegment2D
    _point_class = volmdlr.Point2D

    def __init__(self, points: List[volmdlr.Point2D],
                 triangles: List[Tuple[int, int, int]],
                 edges: List[Tuple[int, int]] = None,
                 name: str = ''):
        DisplayMesh.__init__(self, points, triangles, name=name)

    def area(self):
        """
        Return the area as the sum of areas of triangles
        """
        area = 0.
        for (n1, n2, n3) in self.triangles:
            p1 = self.points[n1]
            p2 = self.points[n2]
            p3 = self.points[n3]
            area += 0.5 * abs((p2 - p1).cross(p3 - p1))
        return area


class DisplayMesh3D(DisplayMesh):
    _linesegment_class = volmdlr.edges.LineSegment3D
    _point_class = volmdlr.Point3D

    def __init__(self, points: List[volmdlr.Point3D],
                 triangles: List[Tuple[int, int, int]], name=''):
        DisplayMesh.__init__(self, points, triangles)

    def to_babylon(self):
        """
        return mesh in babylon format: https://doc.babylonjs.com/how_to/custom
        """
        positions = []
        for p in self.points:
            positions.extend(list(round(p, 6)))

        flatten_indices = []
        for i in self.triangles:
            flatten_indices.extend(i)
        return positions, flatten_indices

    def to_stl(self):
        '''
        Exports to STL
        '''
        # TODO: remove this in the future
        import volmdlr.stl as vmstl
        stl = vmstl.Stl.from_display_mesh(self)
        return stl



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Oct 20 2022

@author: s.bendjebla
"""

# %% Librairies

import volmdlr.wires
import volmdlr.edges

# %% Contour2d_1

primitives = [
    volmdlr.edges.LineSegment2D(volmdlr.Point2D(0.001, 0.014),
                                volmdlr.Point2D(0.001, 0.0125)),

    volmdlr.edges.Arc2D(volmdlr.Point2D(0.001, 0.0125), 
                        volmdlr.Point2D(0.009862829911410362, 0.007744326060968065),
                        volmdlr.Point2D(0.012539936203984454, 0.0)),

    volmdlr.edges.Arc2D(volmdlr.Point2D(0.012539936203984454, 0.0), 
                        volmdlr.Point2D(0.0, -0.012539936203984454),
                        volmdlr.Point2D(-0.012539936203984454, 0.0)),

    volmdlr.edges.Arc2D(volmdlr.Point2D(-0.012539936203984454, 0.0), 
                        volmdlr.Point2D(-0.00921384654213387, 0.008506176103162205),
                        volmdlr.Point2D(-0.001, 0.0125)),

    volmdlr.edges.LineSegment2D(volmdlr.Point2D(-0.001, 0.0125),
                                volmdlr.Point2D(-0.001, 0.014)),

    volmdlr.edges.LineSegment2D(volmdlr.Point2D(-0.001, 0.014),
                                volmdlr.Point2D(0.001, 0.014))
    ]

Contour2d_1 = volmdlr.wires.Contour2D(primitives)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Common primitives 3D
"""

import math

from typing import Tuple, List, Dict
from scipy.optimize import minimize, NonlinearConstraint

import numpy as npy
import matplotlib.pyplot as plt

import dessia_common as dc
import volmdlr
import volmdlr.core
import volmdlr.primitives
import volmdlr.faces
import volmdlr.edges
import volmdlr.wires

# import dessia_common.typings as dct

npy.seterr(divide='raise')


class OpenRoundedLineSegments3D(volmdlr.wires.Wire3D,
                                volmdlr.primitives.RoundedLineSegments):
    _non_eq_attributes = ['name']
    _non_hash_attributes = ['name']
    _generic_eq = True

    line_class = volmdlr.edges.LineSegment3D
    arc_class = volmdlr.edges.Arc3D

    def __init__(self, points: List[volmdlr.Point3D], radius: Dict[str, float],
                 adapt_radius: bool = False, name: str = ''):
        volmdlr.primitives.RoundedLineSegments.__init__(
            self, points, radius, closed=False, adapt_radius=adapt_radius,
            name='')

        volmdlr.wires.Wire3D.__init__(self, self._primitives(), name)

    def arc_features(self, ipoint):
        radius = self.radius[ipoint]
        if self.closed:
            if ipoint == 0:
                pt1 = self.points[-1]
            else:
                pt1 = self.points[ipoint - 1]
            pti = self.points[ipoint]
            if ipoint < self.npoints - 1:
                pt2 = self.points[ipoint + 1]
            else:
                pt2 = self.points[0]
        else:
            pt1 = self.points[ipoint - 1]
            pti = self.points[ipoint]
            pt2 = self.points[ipoint + 1]

        dist1 = (pt1 - pti).norm()
        dist2 = (pt2 - pti).norm()
        dist3 = (pt1 - pt2).norm()
        alpha = math.acos(-(dist3**2 - dist1**2 - dist2**2) / (2 * dist1 * dist2)) / 2.
        dist = radius / math.tan(alpha)

        u1 = (pt1 - pti) / dist1
        u2 = (pt2 - pti) / dist2

        p3 = pti + u1 * dist
        p4 = pti + u2 * dist

        n = u1.cross(u2)
        n /= n.norm()
        v1 = u1.cross(n)
        v2 = u2.cross(n)

        l1 = volmdlr.edges.Line3D(p3, p3 + v1)
        l2 = volmdlr.edges.Line3D(p4, p4 + v2)
        c, _ = l1.minimum_distance_points(l2)

        u3 = u1 + u2  # mean of v1 and v2
        u3 /= u3.norm()

        interior = c - u3 * radius
        return p3, interior, p4, dist, alpha

    def rotation(self, center: volmdlr.Point3D, axis: volmdlr.Vector3D,
                 angle: float):
        """
        OpenRoundedLineSegments3D rotation
        :param center: rotation center
        :param axis: rotation axis
        :param angle: angle rotation
        :return: a new rotated OpenRoundedLineSegments3D
        """
        return self.__class__([point.rotation(center, axis, angle)
                               for point in self.points],
                              self.radius, self.closed, self.name)

    def rotation_inplace(self, center: volmdlr.Point3D,
                         axis: volmdlr.Vector3D,
                         angle: float):
        """
        OpenRoundedLineSegments3D rotation. Object is updated inplace
        :param center: rotation center
        :param axis: rotation axis
        :param angle: rotation angle
        """
        for point in self.points:
            point.rotation_inplace(center, axis, angle)

    def translation(self, offset: volmdlr.Vector3D):
        """
        OpenRoundedLineSegments3D translation
        :param offset: translation vector
        :return: A new translated OpenRoundedLineSegments3D
        """
        return self.__class__([point.translation(offset)
                               for point in self.points],
                              self.radius, self.closed, self.name)

    def translation_inplace(self, offset: volmdlr.Vector3D):
        """
        OpenRoundedLineSegments3D translation. Object is updated inplace
        :param offset: translation vector
        """
        for point in self.points:
            point.translation_inplace(offset)


class ClosedRoundedLineSegments3D(volmdlr.wires.Contour3D,
                                  OpenRoundedLineSegments3D):
    """
    :param points: Points used to draw the wire
    :type points: List of Point3D
    :param radius: Radius used to connect different parts of the wire
    :type radius: {position1(n): float which is the radius linked the n-1 and
    n+1 points, position2(n+1):...}
    """
    _non_serializable_attributes = []
    _non_eq_attributes = ['name']
    _non_hash_attributes = ['name']
    _generic_eq = True

    def __init__(self, points, radius, adapt_radius=False, name=''):
        volmdlr.primitives.RoundedLineSegments.__init__(
                self, points, radius, 'volmdlr.edges.LineSegment3D',
                'volmdlr.edges.Arc3D', closed=True, adapt_radius=adapt_radius,
                name='')

        volmdlr.wires.Wire3D.__init__(self, self._primitives(), name)


class Block(volmdlr.faces.ClosedShell3D):
    _standalone_in_db = True
    _generic_eq = True
    _non_serializable_attributes = ['size', 'bounding_box', 'faces', 'contours', 'plane', 'points', 'polygon2D']
    _non_eq_attributes = ['name', 'color', 'alpha', 'size', 'bounding_box',
                          'faces', 'contours', 'plane', 'points', 'polygon2D']
    _non_hash_attributes = []

    """
    Creates a block
    :param frame: a frame 3D. The origin of the frame is the center of the block,
     the 3 vectors are defining the edges. The frame has not to be orthogonal
    """

    def __init__(self, frame: volmdlr.Frame3D, *,
                 color: Tuple[float, float, float] = None, alpha: float = 1.,
                 name: str = ''):
        self.frame = frame
        self.size = (self.frame.u.norm(),
                     self.frame.v.norm(),
                     self.frame.w.norm())

        faces = self.shell_faces()
        volmdlr.faces.OpenShell3D.__init__(self, faces, color=color,
                                           alpha=alpha, name=name)

    # def __hash__(self):
    #     return hash(self.frame)

    def to_dict(self, use_pointers: bool = False, memo=None, path: str = '#'):
        """
        Custom to_dict for performance
        """
        dict_ = dc.DessiaObject.base_dict(self)
        dict_.update({'color': self.color,
                      'alpha': self.alpha,
                      'frame': self.frame.to_dict()})

        return dict_

    def volume(self):
        return self.size[0] * self.size[1] * self.size[2]

    @classmethod
    def from_bounding_box(cls, bounding_box):
        bb = bounding_box
        xmin, xmax, ymin, ymax, zmin, zmax = bb.xmin, bb.xmax,\
            bb.ymin, bb.ymax, bb.zmin, bb.zmax
        origin = bb.center
        sx, sy, sz = xmax - xmin, ymax - ymin, zmax - zmin
        frame = volmdlr.Frame3D(origin, sx * volmdlr.Vector3D(1, 0, 0),
                                sy * volmdlr.Vector3D(0, 1, 0),
                                sz * volmdlr.Vector3D(0, 0, 1))
        return cls(frame=frame)

    def vertices(self):
        return [self.frame.origin - 0.5 * self.frame.u - 0.5 * self.frame.v - 0.5 * self.frame.w,
                self.frame.origin - 0.5 * self.frame.u + 0.5 * self.frame.v - 0.5 * self.frame.w,
                self.frame.origin + 0.5 * self.frame.u + 0.5 * self.frame.v - 0.5 * self.frame.w,
                self.frame.origin + 0.5 * self.frame.u - 0.5 * self.frame.v - 0.5 * self.frame.w,
                self.frame.origin - 0.5 * self.frame.u - 0.5 * self.frame.v + 0.5 * self.frame.w,
                self.frame.origin - 0.5 * self.frame.u + 0.5 * self.frame.v + 0.5 * self.frame.w,
                self.frame.origin + 0.5 * self.frame.u + 0.5 * self.frame.v + 0.5 * self.frame.w,
                self.frame.origin + 0.5 * self.frame.u - 0.5 * self.frame.v + 0.5 * self.frame.w]

    def edges(self):
        p1, p2, p3, p4, p5, p6, p7, p8 = self.vertices()
        return [volmdlr.edges.LineSegment3D(p1.copy(), p2.copy()),
                volmdlr.edges.LineSegment3D(p2.copy(), p3.copy()),
                volmdlr.edges.LineSegment3D(p3.copy(), p4.copy()),
                volmdlr.edges.LineSegment3D(p4.copy(), p1.copy()),
                volmdlr.edges.LineSegment3D(p5.copy(), p6.copy()),
                volmdlr.edges.LineSegment3D(p6.copy(), p7.copy()),
                volmdlr.edges.LineSegment3D(p7.copy(), p8.copy()),
                volmdlr.edges.LineSegment3D(p8.copy(), p5.copy()),
                volmdlr.edges.LineSegment3D(p1.copy(), p5.copy()),
                volmdlr.edges.LineSegment3D(p2.copy(), p6.copy()),
                volmdlr.edges.LineSegment3D(p3.copy(), p7.copy()),
                volmdlr.edges.LineSegment3D(p4.copy(), p8.copy())]

    def face_contours3d(self):
        e1, e2, e3, e4, e5, e6, e7, e8, e9, e10, e11, e12 = self.edges()
        e5_switch = e5.reverse()
        e6_switch = e6.reverse()
        e7_switch = e7.reverse()
        e8_switch = e8.reverse()
        e9_switch = e9.reverse()
        e10_switch = e10.reverse()
        e11_switch = e11.reverse()
        e12_switch = e12.reverse()
        return [volmdlr.wires.Contour3D([e1.copy(), e2.copy(), e3.copy(), e4.copy()]),
                volmdlr.wires.Contour3D([e5.copy(), e6.copy(), e7.copy(), e8.copy()]),
                # volmdlr.Contour3D([e1.copy(), e9.copy(), e5.copy(), e10.copy()]),
                volmdlr.wires.Contour3D([e1.copy(), e10.copy(), e5_switch.copy(), e9_switch.copy()]),
                # volmdlr.Contour3D([e2.copy(), e10.copy(), e6.copy(), e11.copy()]),
                volmdlr.wires.Contour3D([e2.copy(), e11.copy(), e6_switch.copy(), e10_switch.copy()]),
                # volmdlr.Contour3D([e3.copy(), e11.copy(), e7.copy(), e12.copy()]),
                volmdlr.wires.Contour3D([e3.copy(), e12.copy(), e7_switch.copy(), e11_switch.copy()]),
                # volmdlr.Contour3D([e4.copy(), e12.copy(), e8.copy(), e9.copy()])]
                volmdlr.wires.Contour3D([e4.copy(), e9.copy(), e8_switch.copy(), e12_switch.copy()])]

    def shell_faces(self):
        hlx = 0.5 * self.frame.u.norm()
        hly = 0.5 * self.frame.v.norm()
        hlz = 0.5 * self.frame.w.norm()
        frame = self.frame.copy()
        frame.normalize()
        xm_frame = volmdlr.Frame3D(frame.origin - 0.5 * self.frame.u,
                                   frame.v, frame.w, frame.u)
        xp_frame = volmdlr.Frame3D(frame.origin + 0.5 * self.frame.u,
                                   frame.v, frame.w, frame.u)
        ym_frame = volmdlr.Frame3D(frame.origin - 0.5 * self.frame.v,
                                   frame.w, frame.u, frame.v)
        yp_frame = volmdlr.Frame3D(frame.origin + 0.5 * self.frame.v,
                                   frame.w, frame.u, frame.v)
        zm_frame = volmdlr.Frame3D(frame.origin - 0.5 * self.frame.w,
                                   frame.u, frame.v, frame.w)
        zp_frame = volmdlr.Frame3D(frame.origin + 0.5 * self.frame.w,
                                   frame.u, frame.v, frame.w)

        xm_face = volmdlr.faces.Plane3D(xm_frame)\
            .rectangular_cut(-hly, hly, -hlz, hlz)
        xp_face = volmdlr.faces.Plane3D(xp_frame)\
            .rectangular_cut(-hly, hly, -hlz, hlz)
        ym_face = volmdlr.faces.Plane3D(ym_frame)\
            .rectangular_cut(-hlz, hlz, -hlx, hlx)
        yp_face = volmdlr.faces.Plane3D(yp_frame)\
            .rectangular_cut(-hlz, hlz, -hlx, hlx)
        zm_face = volmdlr.faces.Plane3D(zm_frame)\
            .rectangular_cut(-hlx, hlx, -hly, hly)
        zp_face = volmdlr.faces.Plane3D(zp_frame)\
            .rectangular_cut(-hlx, hlx, -hly, hly)

        return [xm_face, xp_face, ym_face, yp_face, zm_face, zp_face]

    def rotation(self, center: volmdlr.Point3D, axis: volmdlr.Vector3D,
                 angle: float):
        """
        Block rotation
        :param center: rotation center
        :param axis: rotation axis
        :param angle: angle rotation
        :return: a new rotated Block
        """
        new_frame = self.frame.rotation(center, axis, angle)
        return Block(new_frame, color=self.color,
                     alpha=self.alpha, name=self.name)

    def rotation_inplace(self, center: volmdlr.Point3D, axis: volmdlr.Vector3D,
                         angle: float):
        """
        Block rotation. Object is updated inplace
        :param center: rotation center
        :param axis: rotation axis
        :param angle: rotation angle
        """
        self.frame.rotation_inplace(center, axis, angle)
        self.faces = self.shell_faces()

    def translation(self, offset: volmdlr.Vector3D):
        """
        Block translation
        :param offset: translation vector
        :return: A new translated Block
        """
        new_frame = self.frame.translation(offset)
        return Block(new_frame, color=self.color,
                     alpha=self.alpha, name=self.name)

    def translation_inplace(self, offset: volmdlr.Vector3D):
        """
        Block translation. Object is updated inplace
        :param offset: translation vector
        """
        self.frame.translation_inplace(offset)
        self.faces = self.shell_faces()

    def cut_by_orthogonal_plane(self, plane_3d: volmdlr.faces.Plane3D):
        bouding_box = self.bounding_box
        if plane_3d.frame.w.dot(volmdlr.Vector3D(1, 0, 0)) == 0:
            pass
        elif plane_3d.frame.w.dot(volmdlr.Vector3D(0, 1, 0)) == 0:
            pass
        elif plane_3d.frame.w.dot(volmdlr.Vector3D(0, 0, 1)) == 0:
            pass
        else:
            raise KeyError('plane is not orthogonal either with x, y or z')
        dir1 = plane_3d.frame.u
        dir2 = plane_3d.frame.v
        point_min = volmdlr.Point3D(bouding_box.xmin, bouding_box.ymin,
                                    bouding_box.zmin)
        point_max = volmdlr.Point3D(bouding_box.xmax, bouding_box.ymax,
                                    bouding_box.zmax)
        points = [volmdlr.Point2D(point_min.dot(dir1), point_min.dot(dir2)),
                  volmdlr.Point2D(point_min.dot(dir1), point_max.dot(dir2)),
                  volmdlr.Point2D(point_max.dot(dir1), point_max.dot(dir2)),
                  volmdlr.Point2D(point_max.dot(dir1), point_min.dot(dir2))]
        contour_2d = volmdlr.faces.Surface2D(
            volmdlr.wires.ClosedPolygon2D(points), [])

        return volmdlr.faces.PlaneFace3D(plane_3d, contour_2d)

    def frame_mapping_parametres(self, frame: volmdlr.Frame3D,
                                 side: str):
        basis = frame.basis()
        if side == 'new':
            new_origin = frame.new_coordinates(self.frame.origin)
            new_u = basis.new_coordinates(self.frame.u)
            new_v = basis.new_coordinates(self.frame.v)
            new_w = basis.new_coordinates(self.frame.w)
            new_frame = volmdlr.Frame3D(new_origin, new_u, new_v, new_w)
        elif side == 'old':
            new_origin = frame.old_coordinates(self.frame.origin)
            new_u = basis.old_coordinates(self.frame.u)
            new_v = basis.old_coordinates(self.frame.v)
            new_w = basis.old_coordinates(self.frame.w)
            new_frame = volmdlr.Frame3D(new_origin, new_u, new_v, new_w)
        else:
            raise ValueError('side value not valid, please specify'
                             'a correct value: \'old\' or \'new\'')
        return new_frame

    def frame_mapping(self, frame: volmdlr.Frame3D, side: str):
        """
        Changes frame_mapping and return a new Frame3D
        side = 'old' or 'new'
        """
        new_frame = self.frame_mapping_parametres(frame, side)
        return Block(new_frame, color=self.color,
                     alpha=self.alpha, name=self.name)

    def frame_mapping_inplace(self, frame: volmdlr.Frame3D, side: str):
        """
        Changes frame_mapping and the object is updated inplace
        side = 'old' or 'new'
        """
        new_frame = self.frame_mapping_parametres(frame, side)
        self.frame = new_frame
        self.faces = self.shell_faces()

    def copy(self, deep=True, memo=None):
        new_origin = self.frame.origin.copy()
        new_u = self.frame.u.copy()
        new_v = self.frame.v.copy()
        new_w = self.frame.w.copy()
        new_frame = volmdlr.Frame3D(new_origin, new_u, new_v, new_w)
        return Block(new_frame, color=self.color,
                     alpha=self.alpha, name=self.name)

    def plot_data(self, x3d, y3d, marker=None, color='black', stroke_width=1,
                  dash=False, opacity=1, arrow=False):
        lines = []
        for edge3D in self.edges():
            lines.append(edge3D.plot_data(x3d, y3d, marker, color,
                                          stroke_width, dash, opacity, arrow))

        return lines

    def plot2d(self, x3d, y3d, ax=None):
        if ax is None:
            fig, ax = plt.subplots()
            ax.set_aspect('equal')
        else:
            fig = None

        for edge3D in self.edges():
            # edge2D = edge3D.PlaneProjection2D()
            edge3D.plot2d(x3d, y3d, ax=ax)

        return fig, ax


class ExtrudedProfile(volmdlr.faces.ClosedShell3D):
    """
    TODO: In the future change to a frame and a surface2D and an extrusion vector
    """
    _non_serializable_attributes = ['faces', 'inner_contours3d',
                                    'outer_contour3d']

    def __init__(self, plane_origin: volmdlr.Point3D,
                 x: volmdlr.Vector3D, y: volmdlr.Vector3D,
                 outer_contour2d: volmdlr.wires.Contour2D,
                 inner_contours2d: List[volmdlr.wires.Contour2D],
                 extrusion_vector: volmdlr.Vector3D,
                 color: Tuple[float, float, float] = None, alpha: float = 1.,
                 name: str = ''):
        self.plane_origin = plane_origin

        self.outer_contour2d = outer_contour2d
        self.outer_contour3d = outer_contour2d.to_3d(plane_origin, x, y)

        self.inner_contours2d = inner_contours2d
        self.extrusion_vector = extrusion_vector
        self.inner_contours3d = []
        self.x = x
        self.y = y
        self.color = color

        bool_areas = []
        for contour in inner_contours2d:
            self.inner_contours3d.append(contour.to_3d(plane_origin, x, y))
            if contour.area() > outer_contour2d.area():
                bool_areas.append(True)
            else:
                bool_areas.append(False)
        if any(bool_areas):
            raise ValueError('At least one inner contour is not contained in outer_contour.')

        faces = self.shell_faces()

        volmdlr.faces.ClosedShell3D.__init__(self, faces, color=color,
                                             alpha=alpha, name=name)

    def to_dict(self, use_pointers: bool = False, memo=None, path: str = '#'):
        """

        """
        dict_ = dc.DessiaObject.base_dict(self)
        dict_.update({'color': self.color,
                      'alpha': self.alpha,
                      'plane_origin': self.plane_origin.to_dict(),
                      'outer_contour2d': self.outer_contour2d.to_dict(),
                      'inner_contours2d': [c.to_dict() for c in self.inner_contours2d],
                      'extrusion_vector': self.extrusion_vector.to_dict(),
                      'x': self.x.to_dict(),
                      'y': self.y.to_dict(),
                      })

        return dict_

    def copy(self, deep=True, memo=None):
        return self.__class__(plane_origin=self.plane_origin.copy(),
                              x=self.x.copy(),
                              y=self.y.copy(),
                              outer_contour2d=self.outer_contour2d.copy(),
                              inner_contours2d=[c.copy() for c in self.inner_contours2d],
                              extrusion_vector=self.extrusion_vector.copy(),
                              color=self.color,
                              alpha=self.alpha,
                              name=self.name)

    def shell_faces(self):
        lower_plane = volmdlr.faces.Plane3D.from_plane_vectors(
            self.plane_origin, self.x, self.y)
        lower_face = volmdlr.faces.PlaneFace3D(
            lower_plane, volmdlr.faces.Surface2D(self.outer_contour2d,
                                                 self.inner_contours2d))

        upper_face = lower_face.translation(self.extrusion_vector)
        lateral_faces = []
        for p in self.outer_contour3d.primitives:
            lateral_faces.extend(p.extrusion(self.extrusion_vector))

        for inner_contour in self.inner_contours3d:
            for p in inner_contour.primitives:
                lateral_faces.extend(p.extrusion(self.extrusion_vector))

        return [lower_face] + [upper_face] + lateral_faces

    # def plot(self, ax=None, color:str='k', alpha:float=1):
    #     if ax is None:
    #         fig, ax = plt.subplots()
    #         ax.set_aspect('equal')
    #     for contour in [self.outer_contour2d]+self.inner_contours2d:
    #         for primitive in contour.primitives:
    #             primitive.plot(ax)
    #     ax.margins(0.1)
    #     return ax

    def FreeCADExport(self, ip):
        name = 'primitive' + str(ip)
        s = 'Wo = []\n'
        s += 'Eo = []\n'
        for prim_index, primitive in enumerate(self.outer_contour3d.primitives):
            s += primitive.FreeCADExport('L{}'.format(prim_index))
            s += 'Eo.append(Part.Edge(L{}))\n'.format(ip)
        s += 'Wo.append(Part.Wire(Eo[:]))\n'
        s += 'Fo = Part.Face(Wo)\n'

        s += 'Fi = []\n'
        s += 'W = []\n'
        for ic, contour in enumerate(self.inner_contours3d):
            s += 'E = []\n'
            for primitive_index, primitive in enumerate(contour.primitives):
                s += primitive.FreeCADExport('L{}_{}'.format(ic, primitive_index))
                s += 'E.append(Part.Edge(L{}_{}))\n'.format(ic, primitive_index)
            s += 'Wi = Part.Wire(E[:])\n'
            s += 'Fi.append(Part.Face(Wi))\n'

        if len(self.inner_contours3d) != 0:
            s += 'Fo = Fo.cut(Fi)\n'
        e1, e2, e3 = round(1000 * self.extrusion_vector, 6)

        s += '{} = Fo.extrude(fc.Vector({}, {}, {}))\n'.format(name, e1,
                                                               e2, e3)
        return s

    def area(self):
        areas = self.outer_contour2d.area()
        areas -= sum([contour.area() for contour in self.inner_contours2d])

        # sic=list(npy.argsort(areas))[::-1]# sorted indices of contours
        # area=areas[sic[0]]

        # for i in sic[1:]:
        #     area-=self.contours2D[i].Area()
        return areas

    def volume(self):
        z = self.x.cross(self.y)
        z.normalize()
        return self.area() * self.extrusion_vector.dot(z)

    def frame_mapping_parameters(self, frame: volmdlr.Frame3D,
                                 side: str):
        basis = frame.basis()
        if side == 'old':
            extrusion_vector = basis.old_coordinates(self.extrusion_vector)
            x = basis.old_coordinates(self.x)
            y = basis.old_coordinates(self.y)
        elif side == 'new':
            extrusion_vector = basis.new_coordinates(self.extrusion_vector)
            x = basis.new_coordinates(self.x)
            y = basis.new_coordinates(self.y)
        else:
            raise ValueError('side must be either old or new')
        return extrusion_vector, x, y

    def frame_mapping(self, frame: volmdlr.Frame3D, side: str):
        """
        Changes frame_mapping and return a new ExtrudeProfile
        side = 'old' or 'new'
        """
        extrusion_vector, x, y = self.frame_mapping_parameters(frame,
                                                               side)
        return ExtrudedProfile(
            self.plane_origin.frame_mapping(frame, side),
            x, y, self.outer_contour2d, self.inner_contours2d,
            extrusion_vector)

    def frame_mapping_inplace(self, frame: volmdlr.Frame3D, side: str):
        """
        Changes frame_mapping and the object is updated inplace
        side = 'old' or 'new'
        """
        self.extrusion_vector, self.x, self.y =\
            self.frame_mapping_parameters(frame, side)
        self.plane_origin.frame_mapping_inplace(frame, side)

    def rotation(self, center: volmdlr.Point3D, axis: volmdlr.Vector3D,
                 angle: float):
        """
        ExtrudedProfile rotation
        :param center: rotation center
        :param axis: rotation axis
        :param angle: angle rotation
        :return: a new rotated ExtrudedProfile
        """
        return self.__class__(
            plane_origin=self.plane_origin.rotation(center, axis, angle),
            x=self.x.rotation(volmdlr.O3D, axis, angle),
            y=self.y.rotation(volmdlr.O3D, axis, angle),
            outer_contour2d=self.outer_contour2d,
            inner_contours2d=self.inner_contours2d,
            extrusion_vector=self.extrusion_vector.rotation(volmdlr.O3D,
                                                            axis, angle),
            color=self.color, alpha=self.alpha)

    def rotation_inplace(self, center: volmdlr.Point3D, axis: volmdlr.Vector3D,
                         angle: float):
        """
        ExtrudedProfile rotation. Object is updated inplace
        :param center: rotation center
        :param axis: rotation axis
        :param angle: rotation angle
        """
        self.plane_origin.rotation_inplace(center, axis, angle)
        self.x.rotation_inplace(volmdlr.O3D, axis, angle)
        self.y.rotation_inplace(volmdlr.O3D, axis, angle)
        self.extrusion_vector.rotation_inplace(volmdlr.O3D, axis, angle)

    def translation(self, offset: volmdlr.Vector3D):
        """
        ExtrudedProfile translation
        :param offset: translation vector
        :return: A new translated ExtrudedProfile
        """
        return self.__class__(
            plane_origin=self.plane_origin.translation(offset),
            x=self.x, y=self.y,
            outer_contour2d=self.outer_contour2d,
            inner_contours2d=self.inner_contours2d,
            extrusion_vector=self.extrusion_vector,
            color=self.color, alpha=self.alpha)

    def translation_inplace(self, offset: volmdlr.Vector3D):
        """
        ExtrudedProfile translation. Object is updated inplace
        :param offset: translation vector
        """
        self.plane_origin.translation_inplace(offset)


class RevolvedProfile(volmdlr.faces.ClosedShell3D):
    """

    """
    _non_serializable_attributes = ['faces', 'contour3D']

    def __init__(self, plane_origin: volmdlr.Point3D,
                 x: volmdlr.Vector3D, y: volmdlr.Vector3D,
                 contour2d: volmdlr.wires.Contour2D,
                 axis_point: volmdlr.Point3D, axis: volmdlr.Vector3D,
                 angle: float = 2 * math.pi, *,
                 color: Tuple[float, float, float] = None, alpha: float = 1,
                 name: str = ''):

        self.contour2d = contour2d
        self.axis_point = axis_point
        self.axis = axis
        self.angle = angle
        self.plane_origin = plane_origin
        self.x = x
        self.y = y
        self.contour3d = self.contour2d.to_3d(plane_origin, x, y)

        faces = self.shell_faces()
        volmdlr.faces.ClosedShell3D.__init__(self, faces, color=color,
                                             alpha=alpha, name=name)

    def to_dict(self, use_pointers: bool = False, memo=None, path: str = '#'):
        """
        Custom to dict for perf
        """
        dict_ = dc.DessiaObject.base_dict(self)
        dict_.update({'color': self.color,
                      'alpha': self.alpha,
                      'plane_origin': self.plane_origin.to_dict(),
                      'contour2d': self.contour2d.to_dict(),
                      'axis_point': self.axis_point.to_dict(),
                      'x': self.x.to_dict(),
                      'y': self.y.to_dict(),
                      'angle': self.angle
                      })

        return dict_

    def copy(self, deep=True, memo=None):
        return self.__class__(plane_origin=self.plane_origin.copy(),
                              x=self.x.copy(), y=self.y.copy(),
                              contour2d=self.contour2d.copy(),
                              axis=self.axis.copy(), angle=self.angle,
                              axis_point=self.axis_point.copy(),
                              color=self.color, alpha=self.alpha,
                              name=self.name)

    def shell_faces(self):
        faces = []

        for edge in self.contour3d.primitives:
            faces.extend(edge.revolution(self.axis_point,
                                         self.axis, self.angle))

        if not math.isclose(self.angle, volmdlr.TWO_PI, abs_tol=1e-9):
            # Adding contours face to close
            w = self.x.cross(self.y)
            plane1 = volmdlr.faces.Plane3D(volmdlr.Frame3D(self.plane_origin,
                                                           self.x,
                                                           self.y,
                                                           w))
            face1 = volmdlr.faces.PlaneFace3D(
                plane1, volmdlr.faces.Surface2D(self.contour2d, []))
            face2 = face1.rotation(self.axis_point, self.axis, self.angle)
            faces.append(face1)
            faces.append(face2)

        return faces

    def FreeCADExport(self, ip, ndigits=3):
        name = 'primitive' + str(ip)
        s = 'W=[]\n'
#        for ic, contour in enumerate(self.contours3D):
        s += 'L=[]\n'
        for ibp, basis_primitive in enumerate(self.contour3d.edges):
            s += basis_primitive.FreeCADExport('L{}_{}'.format(1, ibp), 8)
            s += 'L.append(L{}_{})\n'.format(1, ibp)
        s += 'S = Part.Shape(L)\n'
        s += 'W.append(Part.Wire(S.Edges))\n'
        s += 'F=Part.Face(W)\n'
        a1, a2, a3 = self.axis.vector
        ap1, ap2, ap3 = self.axis_point.vector
        ap1 = round(ap1 * 1000, ndigits)
        ap2 = round(ap2 * 1000, ndigits)
        ap3 = round(ap3 * 1000, ndigits)
        angle = self.angle / math.pi * 180
        s += '{} = F.revolve(fc.Vector({},{},{}), fc.Vector({},{},{}),{})\n'.format(name,
                                                                                    ap1, ap2, ap3, a1, a2, a3, angle)
        return s

    def volume(self):
        """
        Volume from guldin formulae
        """
        p1 = self.axis_point.PlaneProjection3D(self.plane_origin,
                                               self.x, self.y)
        p1_2D = p1.To2D(self.axis_point, self.x, self.y)
        p2_3D = self.axis_point + volmdlr.Point3D(self.axis.vector)
        p2_2D = p2_3D.To2D(self.plane_origin, self.x, self.y)
        axis_2D = volmdlr.edges.Line2D(p1_2D, p2_2D)
        com = self.contour2d.center_of_mass()
        if com is not False:
            rg = axis_2D.point_distance(com)
            return self.angle * rg * self.contour2d.area()
        else:
            return 0

    def rotation(self, center: volmdlr.Point3D, axis: volmdlr.Vector3D,
                 angle: float):
        """
        RevolvedProfile rotation
        :param center: rotation center
        :param axis: rotation axis
        :param angle: angle rotation
        :return: a new rotated RevolvedProfile
        """
        return self.__class__(
            plane_origin=self.plane_origin.rotation(center, axis, angle),
            x=self.x.rotation(center=volmdlr.O3D, axis=axis, angle=angle),
            y=self.y.rotation(center=volmdlr.O3D, axis=axis, angle=angle),
            contour2d=self.contour2d,
            axis_point=self.axis_point.rotation(center, axis, angle),
            axis=self.axis.rotation(center=volmdlr.O3D, axis=axis,
                                    angle=angle),
            angle=self.angle,
            color=self.color, alpha=self.alpha)

    def rotation_inplace(self, center: volmdlr.Point3D, axis: volmdlr.Vector3D,
                         angle: float):
        """
        RevolvedProfile rotation. Object is updated inplace
        :param center: rotation center
        :param axis: rotation axis
        :param angle: rotation angle
        """
        self.plane_origin.rotation_inplace(center, axis, angle)
        self.x.rotation_inplace(center=volmdlr.O3D, axis=axis, angle=angle)
        self.y.rotation_inplace(center=volmdlr.O3D, axis=axis, angle=angle)
        self.axis_point.rotation_inplace(center, axis, angle)
        self.axis.rotation_inplace(volmdlr.O3D, axis, angle)

    def translation(self, offset: volmdlr.Vector3D):
        """
        RevolvedProfile translation
        :param offset: translation vector
        :return: A new translated RevolvedProfile
        """
        return self.__class__(
            plane_origin=self.plane_origin.translation(offset),
            x=self.x, y=self.y, contour2d=self.contour2d,
            axis_point=self.axis_point.translation(offset),
            axis=self.axis,
            angle=self.angle,
            color=self.color, alpha=self.alpha)

    def translation_inplace(self, offset: volmdlr.Vector3D):
        """
        RevolvedProfile translation. Object is updated inplace
        :param offset: translation vector
        """
        self.plane_origin.translation_inplace(offset)
        self.axis_point.translation_inplace(offset)

    def frame_mapping_parameters(self, frame: volmdlr.Frame3D,
                                 side: str):
        basis = frame.Basis()
        if side == 'old':
            axis = basis.old_coordinates(self.axis)
            x = basis.old_coordinates(self.x)
            y = basis.old_coordinates(self.y)
        elif side == 'new':
            axis = basis.new_coordinates(self.axis)
            x = basis.new_coordinates(self.x)
            y = basis.new_coordinates(self.y)
        else:
            raise ValueError('side must be either old or new')

        return axis, x, y

    def frame_mapping(self, frame: volmdlr.Frame3D, side: str):
        """
        Changes frame_mapping and return a new RevolvedProfile
        side = 'old' or 'new'
        """
        axis, x, y = self.frame_mapping_parameters(frame, side)
        return RevolvedProfile(
            self.plane_origin.frame_mapping(frame, side),
            x, y, self.contour2d,
            self.axis_point.frame_mapping(frame, side),
            axis, self.angle)

    def frame_mapping_inplace(self, frame: volmdlr.Frame3D, side: str):
        """
        Changes frame_mapping and the object is updated inplace
        side = 'old' or 'new'
        """
        self.axis, self.x, self.y = self.frame_mapping_parameters(frame, side)
        self.plane_origin.frame_mapping_inplace(frame, side)
        self.axis_point.frame_mapping_inplace(frame, side)


class Cylinder(RevolvedProfile):
    """
    Creates a full cylinder with the position, the axis of revolution,
    the radius and the length.
    """

    def __init__(self, position: volmdlr.Point3D, axis: volmdlr.Vector3D,
                 radius: float, length: float,
                 color: Tuple[float, float, float] = None, alpha: float = 1.,
                 name: str = ''):

        self.position = position
        axis.normalize()
        self.axis = axis
        self.radius = radius
        self.length = length
        self.bounding_box = self._bounding_box()

        # Revolved Profile
        p1 = volmdlr.Point2D(-0.5 * self.length, 0.)
        p2 = volmdlr.Point2D(0.5 * self.length, 0.)
        p3 = volmdlr.Point2D(0.5 * self.length, self.radius)
        p4 = volmdlr.Point2D(-0.5 * self.length, self.radius)
        l1 = volmdlr.edges.LineSegment2D(p1, p2)
        l2 = volmdlr.edges.LineSegment2D(p2, p3)
        l3 = volmdlr.edges.LineSegment2D(p3, p4)
        l4 = volmdlr.edges.LineSegment2D(p4, p1)
        contour = volmdlr.wires.Contour2D([l1, l2, l3, l4])
        y = axis.random_unit_normal_vector()
        RevolvedProfile.__init__(self, position, axis, y, contour, position,
                                 axis, color=color, alpha=alpha, name=name)

    def _bounding_box(self):
        """
        This was copied for HollowCylinder. Inheritence removed to avoid problems
        """

        radius = self.radius

        pointA = self.position - self.length / 2 * self.axis
        pointB = self.position + self.length / 2 * self.axis

        dx2 = (pointA[0] - pointB[0])**2
        dy2 = (pointA[1] - pointB[1])**2
        dz2 = (pointA[2] - pointB[2])**2

        # kx = ((dy2 + dz2) / (dx2 + dy2 + dz2))**0.5
        # ky = ((dx2 + dz2) / (dx2 + dy2 + dz2))**0.5
        # kz = ((dx2 + dy2) / (dx2 + dy2 + dz2))**0.5

        if pointA[0] > pointB[0]:
            pointA, pointB = pointB, pointA
        xmin = pointA[0] - (((dy2 + dz2) / (dx2 + dy2 + dz2))**0.5) * radius
        xmax = pointB[0] + (((dy2 + dz2) / (dx2 + dy2 + dz2))**0.5) * radius

        if pointA[1] > pointB[1]:
            pointA, pointB = pointB, pointA
        ymin = pointA[1] - (((dx2 + dz2) / (dx2 + dy2 + dz2))**0.5) * radius
        ymax = pointB[1] + (((dx2 + dz2) / (dx2 + dy2 + dz2))**0.5) * radius

        if pointA[2] > pointB[2]:
            pointA, pointB = pointB, pointA
        zmin = pointA[2] - (((dx2 + dy2) / (dx2 + dy2 + dz2))**0.5) * radius
        zmax = pointB[2] + (((dx2 + dy2) / (dx2 + dy2 + dz2))**0.5) * radius

        return volmdlr.core.BoundingBox(xmin, xmax, ymin, ymax, zmin, zmax)

    def volume(self):
        return self.length * math.pi * self.radius**2

    @classmethod
    def from_extremal_points(cls, point1: volmdlr.Point3D, point2: volmdlr.Point3D,
                             radius: float,
                             color: Tuple[float, float, float] = None, alpha: float = 1,
                             name: str = ''):
        position = 0.5 * (point1 + point2)
        length = point1.point_distance(point2)
        axis = point2 - point1
        axis.normalize()
        return cls(position, axis, radius, length=length,
                   color=color, alpha=alpha, name=name)

    def FreeCADExport(self, ip):
        if self.radius > 0:
            name = 'primitive' + str(ip)
            e = str(1000 * self.length)
            r = str(1000 * self.radius)
            position = 1000 * (self.position - self.axis * self.length / 2.)
            x, y, z = position
            x = str(x)
            y = str(y)
            z = str(z)

            ax, ay, az = self.axis
            ax = str(ax)
            ay = str(ay)
            az = str(az)
            return name + '=Part.makeCylinder(' + r + ',' + e + ',fc.Vector(' + x + ',' + y + \
                ',' + z + '),fc.Vector(' + ax + ',' + ay + ',' + az + '),360)\n'
        else:
            return ''

    def babylon_script(self, name='primitive_mesh'):
        normal_vector1 = self.axis.RandomUnitnormalVector()
        p1 = volmdlr.Point2D((-0.5 * self.length, self.radius))
        p2 = volmdlr.Point2D((0.5 * self.length, self.radius))
        p3 = volmdlr.Point2D((0.5 * self.length, 0.))
        p4 = volmdlr.Point2D((-0.5 * self.length, 0.))
        l1 = volmdlr.edges.LineSegment2D(p1, p2)
        l2 = volmdlr.edges.LineSegment2D(p2, p3)
        l3 = volmdlr.edges.LineSegment2D(p3, p4)
        l4 = volmdlr.edges.LineSegment2D(p4, p1)
        extruded_profile = RevolvedProfile(
            self.position, self.axis, normal_vector1,
            volmdlr.wires.Contour2D([l1, l2, l3, l4]),
            self.position, self.axis, name=self.name)
        return extruded_profile.babylon_script(name=name)

    def rotation(self, center: volmdlr.Point3D, axis: volmdlr.Vector3D,
                 angle: float):
        """
        Cylinder rotation
        :param center: rotation center
        :param axis: rotation axis
        :param angle: angle rotation
        :return: a new rotated Cylinder
        """
        return self.__class__(
            position=self.position.rotation(center, axis, angle),
            axis=self.axis.rotation(volmdlr.O3D, axis, angle),
            length=self.length, radius=self.radius)

    def rotation_inplace(self, center: volmdlr.Point3D, axis: volmdlr.Vector3D,
                         angle: float):
        """
        Cylinder rotation. Object is updated inplace
        :param center: rotation center
        :param axis: rotation axis
        :param angle: rotation angle
        """
        self.position.rotation_inplace(center, axis, angle)
        self.axis.rotation_inplace(volmdlr.O3D, axis, angle)

    def translation(self, offset: volmdlr.Vector3D):
        """
        Cylinder translation
        :param offset: translation vector
        :return: A new translated Cylinder
        """
        return self.__class__(
            position=self.position.translation(offset),
            axis=self.axis, length=self.length, radius=self.radius)

    def translation_inplace(self, offset: volmdlr.Vector3D):
        """
        Cylinder translation. Object is updated inplace
        :param offset: translation vector
        """
        self.position.translation_inplace(offset)

    def frame_mapping(self, frame: volmdlr.Frame3D, side: str):
        """
        Changes frame_mapping and return a new Frame3D
        side = 'old' or 'new'
        """
        basis = frame.basis()
        if side == 'old':
            axis = basis.old_coordinates(self.axis)
        elif side == 'new':
            axis = basis.new_coordinates(self.axis)
        else:
            raise ValueError('side must be either old or new')
        return Cylinder(self.position.frame_mapping(frame, side),
                        axis, self.radius, self.length,
                        color=self.color, alpha=self.alpha)

    def frame_mapping_inplace(self, frame: volmdlr.Frame3D, side: str):
        """
        Changes frame_mapping and the object is updated inplace
        side = 'old' or 'new'
        """
        basis = frame.basis()
        if side == 'old':
            axis = basis.old_coordinates(self.axis)
        elif side == 'new':
            axis = basis.new_coordinates(self.axis)
        else:
            raise ValueError('side must be either old or new')
        self.position.frame_mapping_inplace(frame, side)
        self.axis = axis

    def copy(self, deep=True, memo=None):
        new_position = self.position.copy()
        new_axis = self.axis.copy()
        return Cylinder(new_position, new_axis, self.radius, self.length,
                        color=self.color, alpha=self.alpha, name=self.name)

    def min_distance_to_other_cylinder(self, other_cylinder: 'Cylinder'):
        """
        Compute the minimal distance between two volmdlr cylinders

        :param other_cylinder: volmdlr Cylinder
        :return: minimal distance between two 3D cylinders
        """
        # Local frames of cylinders
        frame0 = volmdlr.Frame3D.from_point_and_vector(point=self.position,
                                                       vector=self.axis,
                                                       main_axis=volmdlr.Z3D)
        frame1 = volmdlr.Frame3D.from_point_and_vector(point=other_cylinder.position,
                                                       vector=other_cylinder.axis,
                                                       main_axis=volmdlr.Z3D)

        # Objective function
        def dist_points(x):
            """
            :param x: coords of a point in cylinder 0 local frame, coords of a point in cylinder 1 local frame
            :return: distance between the two points
            """
            point0 = frame0.old_coordinates(volmdlr.Point3D(x[0], x[1], x[2]))
            point1 = frame1.old_coordinates(volmdlr.Point3D(x[3], x[4], x[5]))

            return point0.point_distance(point1)

        # Initial vector
        p0 = frame0.old_coordinates(volmdlr.O3D)
        p1 = frame1.old_coordinates(volmdlr.O3D)
        x0 = (p0.x, p0.y, p0.z, p1.x, p1.y, p1.z)

        # Constraints
        def constraint_radius_0(x):
            # radius of cylinder 0
            return x[0] ** 2 + x[1] ** 2

        def constraint_height_0(x):
            # height of cylinder 0
            return x[2]

        def constraint_radius_1(x):
            # radius of cylinder 1
            return x[3] ** 2 + x[4] ** 2

        def constraint_height_1(x):
            # height of cylinder 1
            return x[5]

        constraints = [
            NonlinearConstraint(fun=constraint_radius_0, lb=0, ub=self.radius ** 2),
            NonlinearConstraint(fun=constraint_height_0, lb=-self.length / 2, ub=self.length / 2),
            NonlinearConstraint(fun=constraint_radius_1, lb=0, ub=other_cylinder.radius ** 2),
            NonlinearConstraint(fun=constraint_height_1, lb=-other_cylinder.length / 2, ub=other_cylinder.length / 2)
        ]

        return minimize(fun=dist_points, x0=x0, constraints=constraints).fun

    def is_intersecting_other_cylinder(self, other_cylinder: 'Cylinder'):
        """
        :param other_cylinder: volmdlr Cylinder
        :return: boolean, True if cylinders are intersecting, False otherwise
        """
        dist = self.min_distance_to_other_cylinder(other_cylinder)

        return dist < 1e-5


class Cone(RevolvedProfile):
    def __init__(self, position: volmdlr.Point3D, axis: volmdlr.Vector3D,
                 radius: float, length: float,
                 color: Tuple[float, float, float] = None, alpha: float = 1.,
                 name: str = ''):

        self.position = position
        axis.normalize()
        self.axis = axis
        self.radius = radius
        self.length = length
        self.bounding_box = self._bounding_box()

        # Revolved Profile
        p1 = volmdlr.Point2D(0., 0.)
        p2 = volmdlr.Point2D(0., self.radius)
        p3 = volmdlr.Point2D(self.length, 0.)
        l1 = volmdlr.edges.LineSegment2D(p1, p2)
        l2 = volmdlr.edges.LineSegment2D(p2, p3)
        l3 = volmdlr.edges.LineSegment2D(p3, p1)
        contour = volmdlr.wires.Contour2D([l1, l2, l3])
        y = axis.random_unit_normal_vector()
        RevolvedProfile.__init__(self, position, axis, y, contour, position,
                                 axis, color=color, alpha=alpha, name=name)

    def _bounding_box(self):
        """
        A is the point at the basis
        B is the top
        """
        pointA = self.position - self.length / 2 * self.axis
        pointB = self.position + self.length / 2 * self.axis

        dx2 = (pointA[0] - pointB[0])**2
        dy2 = (pointA[1] - pointB[1])**2
        dz2 = (pointA[2] - pointB[2])**2

        # kx = ((dy2 + dz2) / (dx2 + dy2 + dz2))**0.5
        # ky = ((dx2 + dz2) / (dx2 + dy2 + dz2))**0.5
        # kz = ((dx2 + dy2) / (dx2 + dy2 + dz2))**0.5

        x_bound = (pointA[0] - (((dy2 + dz2) / (dx2 + dy2 + dz2))**0.5) * self.radius,
                   pointA[0] + (((dy2 + dz2) / (dx2 + dy2 + dz2))**0.5) * self.radius, pointB[0])
        xmin = min(x_bound)
        xmax = max(x_bound)

        y_bound = (pointA[1] - (((dx2 + dz2) / (dx2 + dy2 + dz2))**0.5) * self.radius,
                   pointA[1] + (((dx2 + dz2) / (dx2 + dy2 + dz2))**0.5) * self.radius, pointB[1])
        ymin = min(y_bound)
        ymax = max(y_bound)

        z_bound = (pointA[2] - (((dx2 + dy2) / (dx2 + dy2 + dz2))**0.5) * self.radius,
                   pointA[2] + (((dx2 + dy2) / (dx2 + dy2 + dz2))**0.5) * self.radius, pointB[2])
        zmin = min(z_bound)
        zmax = max(z_bound)

        return volmdlr.core.BoundingBox(xmin, xmax, ymin, ymax, zmin, zmax)

    def translation(self, offset: volmdlr.Vector3D):
        """
        Cone translation
        :param offset: translation vector
        :return: A new translated Cone
        """
        return self.__class__(position=self.position.translation(offset),
                              axis=self.axis,
                              radius=self.radius,
                              length=self.length,
                              color=self.color,
                              alpha=self.alpha)

    def translation_inplace(self, offset: volmdlr.Vector3D):
        """
        Plane3D translation. Object is updated inplace
        :param offset: translation vector
        """
        self.position.translation_inplace(offset)

    def rotation(self, center: volmdlr.Point3D, axis: volmdlr.Vector3D,
                 angle: float):
        """
        Cone rotation
        :param center: rotation center
        :param axis: rotation axis
        :param angle: angle rotation
        :return: a new rotated Cone
        """
        return self.__class__(position=self.position.rotation(
            center, axis, angle), axis=self.axis.rotation(center, axis, angle),
            radius=self.radius, length=self.length, color=self.color,
            alpha=self.alpha)

    def rotation_inplace(self, center: volmdlr.Point3D, axis: volmdlr.Vector3D,
                         angle: float):
        """
        Cone rotation. Object is updated inplace
        :param center: rotation center
        :param axis: rotation axis
        :param angle: rotation angle
        """
        self.position.rotation_inplace(center, axis, angle)
        self.axis.rotation_inplace(center, axis, angle)

    def volume(self):
        return self.length * math.pi * self.radius**2 / 3


class HollowCylinder(RevolvedProfile):
    def __init__(self, position: volmdlr.Point3D, axis: volmdlr.Vector3D,
                 inner_radius: float, outer_radius: float, length: float,
                 color: Tuple[float, float, float] = None, alpha: float = 1,
                 name: str = ''):
        volmdlr.core.Primitive3D.__init__(self, name=name)
        self.position = position
        axis.normalize()
        self.axis = axis
        self.inner_radius = inner_radius
        self.outer_radius = outer_radius
        self.length = length

        # Revolved Profile
        p1 = volmdlr.Point2D(-0.5 * self.length, self.inner_radius)
        p2 = volmdlr.Point2D(0.5 * self.length, self.inner_radius)
        p3 = volmdlr.Point2D(0.5 * self.length, self.outer_radius)
        p4 = volmdlr.Point2D(-0.5 * self.length, self.outer_radius)
        l1 = volmdlr.edges.LineSegment2D(p1, p2)
        l2 = volmdlr.edges.LineSegment2D(p2, p3)
        l3 = volmdlr.edges.LineSegment2D(p3, p4)
        l4 = volmdlr.edges.LineSegment2D(p4, p1)
        contour = volmdlr.wires.Contour2D([l1, l2, l3, l4])
        y = axis.random_unit_normal_vector()
        # contour.plot()
        RevolvedProfile.__init__(self, position, axis, y, contour, position,
                                 axis, color=color, alpha=alpha, name=name)

    def _bounding_box(self):

        radius = self.outer_radius

        pointA = self.position - self.length / 2 * self.axis
        pointB = self.position + self.length / 2 * self.axis

        dx2 = (pointA[0] - pointB[0])**2
        dy2 = (pointA[1] - pointB[1])**2
        dz2 = (pointA[2] - pointB[2])**2

        # kx = ((dy2 + dz2) / (dx2 + dy2 + dz2))**0.5
        # ky = ((dx2 + dz2) / (dx2 + dy2 + dz2))**0.5
        # kz = ((dx2 + dy2) / (dx2 + dy2 + dz2))**0.5

        if pointA[0] > pointB[0]:
            pointA, pointB = pointB, pointA
        xmin = pointA[0] - (((dy2 + dz2) / (dx2 + dy2 + dz2))**0.5) * radius
        xmax = pointB[0] + (((dy2 + dz2) / (dx2 + dy2 + dz2))**0.5) * radius

        if pointA[1] > pointB[1]:
            pointA, pointB = pointB, pointA
        ymin = pointA[1] - (((dx2 + dz2) / (dx2 + dy2 + dz2))**0.5) * radius
        ymax = pointB[1] + (((dx2 + dz2) / (dx2 + dy2 + dz2))**0.5) * radius

        if pointA[2] > pointB[2]:
            pointA, pointB = pointB, pointA
        zmin = pointA[2] - (((dx2 + dy2) / (dx2 + dy2 + dz2))**0.5) * radius
        zmax = pointB[2] + (((dx2 + dy2) / (dx2 + dy2 + dz2))**0.5) * radius

        return volmdlr.core.BoundingBox(xmin, xmax, ymin, ymax, zmin, zmax)

    def volume(self):
        return self.length * math.pi * (self.outer_radius**2
                                        - self.inner_radius**2)

    def copy(self):
        new_position = self.position.copy()
        new_axis = self.axis.copy()
        return HollowCylinder(new_position, new_axis, self.inner_radius, self.outer_radius, self.length,
                              color=self.color, alpha=self.alpha, name=self.name)

    @classmethod
    def from_extremal_points(cls, point1: volmdlr.Point3D, point2: volmdlr.Point3D,
                             inner_radius: float, outer_radius: float,
                             color: Tuple[float, float, float] = None, alpha: float = 1,
                             name: str = ''):
        position = 0.5 * (point1 + point2)
        length = point1.point_distance(point2)
        axis = point2 - point1
        axis.normalize()
        return cls(position, axis, inner_radius=inner_radius, outer_radius=outer_radius, length=length,
                   color=color, alpha=alpha, name=name)

    def FreeCADExport(self, ip):
        if self.outer_radius > 0.:
            name = 'primitive' + str(ip)
            re = round(1000 * self.outer_radius, 6)
            ri = round(1000 * self.inner_radius, 6)
            x, y, z = round((1000 * (self.position - self.axis * self.length / 2)),
                            6)
            ax, ay, az = npy.round(self.axis.vector, 6)

            s = 'C2 = Part.makeCircle({}, fc.Vector({}, {}, {}),fc.Vector({}, {}, {}))\n'.format(
                re, x, y, z, ax, ay, az)
            s += 'W2 = Part.Wire(C2.Edges)\n'
            s += 'F2 = Part.Face(W2)\n'

            if self.inner_radius != 0.:
                s += 'C1 = Part.makeCircle({}, fc.Vector({}, {}, {}),fc.Vector({}, {}, {}))\n'.format(ri,
                                                                                                      x, y, z, ax, ay, az)
                s += 'W1 = Part.Wire(C1.Edges)\n'
                s += 'F1 = Part.Face(W1)\n'
                s += 'F2 = F2.cut(F1)\n'

            vx, vy, vz = round(self.axis * self.length * 1000, 6)

            s += '{} = F2.extrude(fc.Vector({}, {}, {}))\n'.format(name, vx,
                                                                   vy, vz)
            return s

        else:
            return ''

    def babylon_script(self, name='primitive_mesh'):
        normal_vector1 = self.axis.RandomUnitnormalVector()
        p1 = volmdlr.Point2D((-0.5 * self.length, self.outer_radius))
        p2 = volmdlr.Point2D((0.5 * self.length, self.outer_radius))
        p3 = volmdlr.Point2D((0.5 * self.length, self.inner_radius))
        p4 = volmdlr.Point2D((-0.5 * self.length, self.inner_radius))
        l1 = volmdlr.edges.LineSegment2D(p1, p2)
        l2 = volmdlr.edges.LineSegment2D(p2, p3)
        l3 = volmdlr.edges.LineSegment2D(p3, p4)
        l4 = volmdlr.edges.LineSegment2D(p4, p1)
        extruded_profile = RevolvedProfile(self.position,
                                           self.axis, normal_vector1,
                                           volmdlr.wires.Contour2D([l1, l2, l3, l4]),
                                           self.position, self.axis,
                                           name=self.name)
        return extruded_profile.babylon_script(name=name)

    def rotation(self, center: volmdlr.Point3D, axis: volmdlr.Vector3D,
                 angle: float):
        """
        HollowCylinder rotation
        :param center: rotation center
        :param axis: rotation axis
        :param angle: angle rotation
        :return: a new rotated HollowCylinder
        """
        return self.__class__(
            position=self.position.rotation(center, axis, angle),
            axis=self.axis.rotation(volmdlr.O3D, axis, angle),
            length=self.length, inner_radius=self.inner_radius,
            outer_radius=self.outer_radius)

    def rotation_inplace(self, center: volmdlr.Point3D, axis: volmdlr.Vector3D,
                         angle: float):
        """
        HollowCylinder rotation. Object is updated inplace
        :param center: rotation center
        :param axis: rotation axis
        :param angle: rotation angle
        """
        self.position.rotation_inplace(center, axis, angle)
        self.axis.rotation_inplace(volmdlr.O3D, axis, angle)

    def translation(self, offset: volmdlr.Vector3D):
        """
        HollowCylinder translation
        :param offset: translation vector
        :return: A new translated HollowCylinder
        """
        return self.__class__(
            position=self.position.translation(offset), axis=self.axis,
            length=self.length, inner_radius=self.inner_radius,
            outer_radius=self.outer_radius)

    def translation_inplace(self, offset: volmdlr.Vector3D):
        """
        HollowCylinder translation. Object is updated inplace
        :param offset: translation vector
        """
        self.position.translation_inplace(offset)

    def frame_mapping(self, frame: volmdlr.Frame3D, side: str):
        """
        Changes frame_mapping and return a new HollowCylinder
        side = 'old' or 'new'
        """
        basis = frame.basis()
        if side == 'old':
            axis = basis.old_coordinates(self.axis)
        elif side == 'new':
            axis = basis.new_coordinates(self.axis)
        else:
            raise ValueError('side must be either old or new')

        return HollowCylinder(
            position=self.position.frame_mapping(frame, side),
            axis=axis, inner_radius=self.inner_radius,
            outer_radius=self.outer_radius, length=self.length)

    def frame_mapping_inplace(self, frame: volmdlr.Frame3D, side: str):
        """
        Changes frame_mapping and the object is updated inplace
        side = 'old' or 'new'
        """
        basis = frame.basis()
        if side == 'old':
            axis = basis.old_coordinates(self.axis)
        elif side == 'new':
            axis = basis.new_coordinates(self.axis)
        else:
            raise ValueError('side must be either old or new')
        self.position.frame_mapping_inplace(frame, side)
        self.axis = axis


class Sweep(volmdlr.faces.ClosedShell3D):
    """
    Sweep a 2D contour along a Wire3D
    """

    def __init__(self, contour2d: List[volmdlr.wires.Contour2D],
                 wire3d: volmdlr.wires.Wire3D, *,
                 color: Tuple[float, float, float] = None, alpha: float = 1,
                 name: str = ''):
        self.contour2d = contour2d
        self.wire3d = wire3d
        self.frames = []

        faces = self.shell_faces()
        volmdlr.faces.ClosedShell3D.__init__(self, faces, color=color,
                                             alpha=alpha, name=name)

    def to_dict(self, use_pointers: bool = False, memo=None, path: str = '#'):
        """
        Custom to dict for perf
        """
        dict_ = dc.DessiaObject.base_dict(self)
        dict_.update({'color': self.color,
                      'alpha': self.alpha,
                      'wire3d': self.wire3d.to_dict(),
                      'contour2d': self.contour2d.to_dict()
                      })

        return dict_

    def shell_faces(self):
        """
        For now it does not take into account rotation of sections
        """

        # End  planar faces
        w = self.wire3d.primitives[0].unit_direction_vector(0.)
        u = self.wire3d.primitives[0].unit_normal_vector(0.)
        if not u:
            u = w.deterministic_unit_normal_vector()
        v = w.cross(u)

        start_plane = volmdlr.faces.Plane3D(
            volmdlr.Frame3D(self.wire3d.point_at_abscissa(0.), u, v, w)
        )

        l_last_primitive = self.wire3d.primitives[-1].length()
        w = self.wire3d.primitives[-1].unit_direction_vector(l_last_primitive)
        u = self.wire3d.primitives[-1].unit_normal_vector(l_last_primitive)
        if not u:
            u = w.deterministic_unit_normal_vector()
        v = w.cross(u)

        end_plane = volmdlr.faces.Plane3D(
            volmdlr.Frame3D(self.wire3d.primitives[-1].point_at_abscissa(
                l_last_primitive),
                            u, v, w))

        faces = [volmdlr.faces.PlaneFace3D(
            start_plane,
            volmdlr.faces.Surface2D(self.contour2d, [])),
                 volmdlr.faces.PlaneFace3D(
            end_plane,
            volmdlr.faces.Surface2D(self.contour2d, []))]

        for wire_primitive in self.wire3d.primitives:
            # tangent, normal = wire_primitive.frenet(0.)
            tangent = wire_primitive.unit_direction_vector(0.)
            normal = wire_primitive.unit_normal_vector(0.)

            if normal is None:
                normal = tangent.deterministic_unit_normal_vector()
            n2 = tangent.cross(normal)
            contour3d = self.contour2d.to_3d(wire_primitive.start, normal, n2)

            if wire_primitive.__class__ is volmdlr.edges.LineSegment3D:
                for contour_primitive in contour3d.primitives:
                    faces.extend(contour_primitive.extrusion(
                        wire_primitive.length()
                        * wire_primitive.unit_direction_vector()))
            elif wire_primitive.__class__ is volmdlr.edges.Arc3D:
                for contour_primitive in contour3d.primitives:
                    faces.extend(contour_primitive.revolution(
                        wire_primitive.center,
                        wire_primitive.normal,
                        wire_primitive.angle))
            elif wire_primitive.__class__ is volmdlr.wires.Circle3D:
                for contour_primitive in contour3d.primitives:
                    faces.extend(contour_primitive.revolution(
                        wire_primitive.center,
                        wire_primitive.normal,
                        volmdlr.TWO_PI))

            elif wire_primitive.__class__ is volmdlr.edges.BSplineCurve3D or \
                    wire_primitive.__class__ is volmdlr.edges.BezierCurve3D:

                tangents = []
                for k, pt in enumerate(wire_primitive.points):
                    position = k / (len(wire_primitive.points) - 1)
                    tangents.append(wire_primitive.tangent(position))

                circles = []
                for pt, tan in zip(wire_primitive.points, tangents):
                    circles.append(volmdlr.wires.Circle3D.from_center_normal(center=pt,
                                                                             normal=tan,
                                                                             radius=self.contour2d.radius))

                polys = [volmdlr.wires.ClosedPolygon3D(c.tessellation_points()) for c in circles]

                size_v, size_u = len(polys[0].points) + 1, len(polys)
                degree_u, degree_v = 3, 3

                points_3d = []
                for poly in polys:
                    points_3d.extend(poly.points)
                    points_3d.append(poly.points[0])

                bezier_surface3d = volmdlr.faces.BezierSurface3D(degree_u,
                                                                 degree_v,
                                                                 points_3d,
                                                                 size_u,
                                                                 size_v)

                outer_contour = volmdlr.wires.Contour2D([volmdlr.edges.LineSegment2D(volmdlr.O2D, volmdlr.X2D),
                                                         volmdlr.edges.LineSegment2D(
                                                             volmdlr.X2D, volmdlr.X2D + volmdlr.Y2D),
                                                         volmdlr.edges.LineSegment2D(
                                                             volmdlr.X2D + volmdlr.Y2D, volmdlr.Y2D),
                                                         volmdlr.edges.LineSegment2D(volmdlr.Y2D, volmdlr.O2D)])
                surf2d = volmdlr.faces.Surface2D(outer_contour, [])

                bsface3d = volmdlr.faces.BSplineFace3D(bezier_surface3d, surf2d)
                faces.append(bsface3d)

            else:
                raise NotImplementedError(
                    'Unimplemented primitive for sweep: {}'
                    .format(wire_primitive.__class__.__name__))

        return faces

    def frame_mapping(self, frame: volmdlr.Frame3D, side: str):
        """
        Changes frame_mapping and return a new Sweep
        side = 'old' or 'new'
        """
        new_wire = self.wire3d.frame_mapping(frame, side)
        return Sweep(self.contour2d, new_wire, color=self.color,
                     alpha=self.alpha, name=self.name)

    def frame_mapping_inplace(self, frame: volmdlr.Frame3D, side: str):
        """
        Changes frame_mapping and the object is updated inplace
        side = 'old' or 'new'
        """
        self.wire3d.frame_mapping_inplace(frame, side)
        for face in self.faces:
            face.frame_mapping_inplace(frame, side)

    def copy(self, deep=True, memo=None):
        new_contour2d = self.contour2d.copy()
        new_wire3d = self.wire3d.copy()
        return Sweep(new_contour2d, new_wire3d, color=self.color,
                     alpha=self.alpha, name=self.name)


# class Sphere(volmdlr.Primitive3D):
class Sphere(RevolvedProfile):
    def __init__(self, center, radius,
                 color: Tuple[float, float, float] = None, alpha: float = 1.,
                 name: str = ''):
        volmdlr.core.Primitive3D.__init__(self, name=name)
        self.center = center
        self.radius = radius
        self.position = center

        # Revolved Profile for complete sphere
        s = volmdlr.Point2D(-self.radius, 0.01 * self.radius)
        i = volmdlr.Point2D(0, 1.01 * self.radius)
        e = volmdlr.Point2D(self.radius, 0.01 * self.radius)  # Not coherent but it works at first, to change !!

        # s = volmdlr.Point2D((-self.radius, 0))
        # i = volmdlr.Point2D(((math.sqrt(2)/2)*self.radius,(math.sqrt(2)/2)*self.radius))
        # e = volmdlr.Point2D(((-math.sqrt(2)/2)*self.radius,(-math.sqrt(2)/2)*self.radius))

        contour = volmdlr.wires.Contour2D([
            volmdlr.edges.Arc2D(s, i, e), volmdlr.edges.LineSegment2D(s, e)])

        axis = volmdlr.X3D
        y = axis.random_unit_normal_vector()
        RevolvedProfile.__init__(self, center, axis, y, contour, center, axis,
                                 color=color, alpha=alpha, name=name)

    def volume(self):
        return 4 / 3 * math.pi * self.radius**3

    def FreeCADExport(self, ip, ndigits=3):
        name = 'primitive' + str(ip)
        r = 1000 * self.radius
        x, y, z = round(1000 * self.center, ndigits)
        return '{} = Part.makeSphere({}, fc.Vector({}, {}, {}))\n'.format(
            name, r, x, y, z)

    def babylon_script(self, name='primitive_mesh'):
        p1 = volmdlr.Point2D((-self.radius, 0))
        p2 = volmdlr.Point2D((0, self.radius))
        p3 = volmdlr.Point2D((self.radius, 0))
        line = volmdlr.edges.LineSegment2D(p1, p3)
        arc = volmdlr.edges.Arc2D(p1, p2, p3)
        extruded_profile = RevolvedProfile(
            self.position, volmdlr.X3D, volmdlr.Y3D,
            volmdlr.wires.Contour2D([line, arc]), self.position, volmdlr.X3D,
            name=self.name)
        return extruded_profile.babylon_script(name=name)

    def frame_mapping(self, frame: volmdlr.Frame3D, side: str):
        """
        Changes frame_mapping and return a new Sphere
        side = 'old' or 'new'
        """
        return Sphere(self.center.frame_mapping(frame, side), self.radius)

    def frame_mapping_inplace(self, frame: volmdlr.Frame3D, side: str):
        """
        Changes frame_mapping and the object is updated inplace
        side = 'old' or 'new'
        """
        self.center.frame_mapping_inplace(frame, side)

    def to_point_skin(self, resolution: float = 1e-3):
        if resolution > 2 * self.radius:
            return []

        theta = 2 * math.asin(resolution / (2 * self.radius))

        nb_floor = int(math.pi / theta) + 1
        rota_theta = [n * theta for n in range(nb_floor)]

        p1 = self.center + volmdlr.X3D * self.radius
        rota_axis = volmdlr.Y3D

        skin_points = []

        for t in rota_theta:
            pt_floor_init = p1.rotation(self.center, rota_axis, t)

            if math.isclose(t, 0, abs_tol=1e-6) or math.isclose(t, math.pi, abs_tol=1e-6):
                skin_points.append(pt_floor_init)

            else:
                center_floor = volmdlr.Point3D(volmdlr.X3D.dot(pt_floor_init),
                                               self.center.y,
                                               self.center.z)

                r_floor = center_floor.point_distance(pt_floor_init)
                theta_floor = resolution / r_floor

                nb_points_floor = int(2 * math.pi / theta_floor) + 1
                rota_theta_floor = [n * theta_floor for n in range(nb_points_floor)]

                if (2 * math.pi - rota_theta_floor[-1]) / theta_floor <= 0.1:
                    rota_theta_floor.pop()

                for tf in rota_theta_floor:
                    pt_floor = pt_floor_init.rotation(center_floor, volmdlr.X3D, tf)
                    skin_points.append(pt_floor)

        return skin_points

    def to_point_in(self, resolution: float = 1e-3):
        in_points = [self.center]
        nb_spheres = int(self.radius / resolution)
        if nb_spheres == 0:
            return in_points

        spheres_radius = [n * resolution for n in range(1, nb_spheres + 1)]

        if (self.radius - spheres_radius[-1]) / resolution <= 0.1:
            spheres_radius.pop()

        for srad in spheres_radius:
            in_sphere = Sphere(self.center, srad)
            in_points.extend(in_sphere.to_point_skin(resolution=resolution))

        return in_points


class Measure3D(volmdlr.edges.Line3D):

    def __init__(self, point1: volmdlr.Point3D, point2: volmdlr.Point3D,
                 color: Tuple[float, float, float] = (1., 0, 0)):
        self.point1, self.point2 = point1, point2
        self.color = color
        self.distance = (point1 - point2).norm()
        self.bounding_box = self._bounding_box()

    # !!! no eq defined!
    def __hash__(self):
        return hash(self.point1) + hash(self.point2)

    def babylon_script(self):
        s = 'var myPoints = [];\n'
        s += 'var point1 = new BABYLON.Vector3({},{},{});\n'.format(
            self.point1.x, self.point1.y, self.point1.z)
        s += 'myPoints.push(point1);\n'
        s += 'var point2 = new BABYLON.Vector3({},{},{});\n'.format(
            self.point2.x, self.point2.y, self.point2.z)
        s += 'myPoints.push(point2);\n'
        s += 'var line = BABYLON.MeshBuilder.CreateLines("lines", {points: myPoints}, scene);\n'
        s += 'line.color = new BABYLON.Color3({}, {}, {});\n'.format(
            self.color[0], self.color[1], self.color[2])
        return s


class BSplineExtrusion(volmdlr.core.Primitive3D):

    def __init__(self, obj, vectorextru: volmdlr.Vector3D, name: str = ''):
        self.obj = obj
        vectorextru.normalize()
        self.vectorextru = vectorextru
        if obj.__class__ is volmdlr.wires.Ellipse3D:
            self.points = obj.tessel_points
        else:
            self.points = obj.points
        self.name = name

    @classmethod
    def from_step(cls, arguments, object_dict):
        name = arguments[0][1:-1]
        if object_dict[arguments[1]].__class__ is volmdlr.wires.Ellipse3D:
            ell = object_dict[arguments[1]]
            vectextru = -object_dict[arguments[2]]
            return cls(ell, vectextru, name)

        elif object_dict[arguments[1]].__class__ is \
                volmdlr.edges.BSplineCurve3D:
            bsplinecurve = object_dict[arguments[1]]
            vectextru = object_dict[arguments[2]]
            return cls(bsplinecurve, vectextru, name)
        else:
            raise NotImplementedError  # a adapter pour les bpsline



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Module containing grid and relative objects
"""

from typing import List
import numpy as npy
from dessia_common import DessiaObject
import volmdlr
import volmdlr.wires


class Grid2D(DessiaObject):

    def __init__(self, lists_points: List[List[volmdlr.Point2D]],
                 direction: List[str],
                 name: str = ''):
        self.lists_points = lists_points
        self.direction = direction
        DessiaObject.__init__(self, name=name)

    def displacement_compared_to(self, initial_grid2d):
        """
        compute the deformation/displacement (dx,dy) of a grid2d based on an another grid2d

        Parameters
        ----------
        grid2d : Grid2D

        """

        points_2d = initial_grid2d.points
        points_2d_deformed = self.points

        displacement = npy.ones(shape=(len(points_2d), 2))  # Grid2D points displacement
        for i in range(0, len(displacement)):
            displacement[i][0] = points_2d_deformed[i][0] - points_2d[i][0]
            displacement[i][1] = points_2d_deformed[i][1] - points_2d[i][1]

        return displacement

    def find_direction_index(self, direction_axis: str):
        """
        find the index of a given direction_axis

        Parameters
        ----------
        direction_axis : str
            'x' OR 'y'

        Returns
        -------
        index : int

        """

        try:
            index = self.direction.index('+' + direction_axis)
        except ValueError:
            index = self.direction.index('-' + direction_axis)

        return index

    @classmethod
    def from_points(cls, points, points_dim_1, direction):
        """
        define a Grid2D given a list of points, number of points along the 1st dimension, and a direction

        Parameters
        ----------
        points : list[volmdlr.Point2D]
        points_dim1 : int
        direction : List[str]

        """

        lists_points = [points[i:i + points_dim_1] for i in range(0, len(points), points_dim_1)]

        return cls(lists_points, direction)

    @classmethod
    def from_properties(cls, x_limits, y_limits, points_nbr,
                        direction=None):
        """
        Define Grid2d based on the given properties

        Parameters
        ----------
        x_limits : tuple(float, float)
            (x_min, x_max)
        y_limits : tuple(float, float)
            (y_min, y_max)
        points_nbr : tuple(int, int)
            number of points along x-axis and y-axis
        direction : list[str]
            it is used to order the generated points

        Returns
        -------
        Grid2d

        """
        if direction is None:
            direction = ['+x', '+y']

        xmin, xmax = x_limits
        ymin, ymax = y_limits
        points_x, points_y = points_nbr

        # points_2d = []
        grid2d = []
        points = []

        if direction == ['+x', '+y']:
            x = npy.linspace(xmin, xmax, points_x)
            y = npy.linspace(ymin, ymax, points_y)

            for yi in y:
                for xi in x:
                    # points_2d.append(volmdlr.Point2D(xi, yi))
                    points.append(volmdlr.Point2D(xi, yi))

                grid2d.append(points)
                points = []

        elif direction == ['-x', '+y']:
            x = npy.linspace(xmax, xmin, points_x)
            y = npy.linspace(ymin, ymax, points_y)

            for yi in y:
                for xi in x:
                    # points_2d.append(volmdlr.Point2D(xi, yi))
                    points.append(volmdlr.Point2D(xi, yi))

                grid2d.append(points)
                points = []

        elif direction == ['+y', '+x']:
            x = npy.linspace(xmin, xmax, points_x)
            y = npy.linspace(ymin, ymax, points_y)

            for xi in x:
                for yi in y:
                    # points_2d.append(volmdlr.Point2D(xi, yi))
                    points.append(volmdlr.Point2D(xi, yi))

                grid2d.append(points)
                points = []

        elif direction == ['-y', '+x']:
            x = npy.linspace(xmin, xmax, points_x)
            y = npy.linspace(ymax, ymin, points_y)

            for xi in x:
                for yi in y:
                    # points_2d.append(volmdlr.Point2D(xi, yi))
                    points.append(volmdlr.Point2D(xi, yi))

                grid2d.append(points)
                points = []

        elif direction == ['+x', '-y']:
            x = npy.linspace(xmin, xmax, points_x)
            y = npy.linspace(ymax, ymin, points_y)

            for yi in y:
                for xi in x:
                    # points_2d.append(volmdlr.Point2D(xi, yi))
                    points.append(volmdlr.Point2D(xi, yi))

                grid2d.append(points)
                points = []

        elif direction == ['-x', '-y']:
            x = npy.linspace(xmax, xmin, points_x)
            y = npy.linspace(ymax, ymin, points_y)

            for yi in y:
                for xi in x:
                    # points_2d.append(volmdlr.Point2D(xi, yi))
                    points.append(volmdlr.Point2D(xi, yi))

                grid2d.append(points)
                points = []

        elif direction == ['+y', '-x']:
            x = npy.linspace(xmax, xmin, points_x)
            y = npy.linspace(ymin, ymax, points_y)

            for xi in x:
                for yi in y:
                    # points_2d.append(volmdlr.Point2D(xi, yi))
                    points.append(volmdlr.Point2D(xi, yi))

                grid2d.append(points)
                points = []

        elif direction == ['-y', '-x']:
            x = npy.linspace(xmax, xmin, points_x)
            y = npy.linspace(ymax, ymin, points_y)

            for xi in x:
                for yi in y:
                    # points_2d.append(volmdlr.Point2D(xi, yi))
                    points.append(volmdlr.Point2D(xi, yi))

                grid2d.append(points)
                points = []

        return cls(lists_points=grid2d, direction=direction)

    def grid_pattern(self):
        """
        define a list of quadrilateral polygons defined based on Grid2d points

        Returns
        -------
        quadrilateral_polygons : list[volmdlr.wires.ClosedPolygon2D]

        """

        quadrilateral_polygons = []
        length_1, length_2 = len(self.lists_points[0]), len(self.lists_points)
        for i in range(0, length_1 - 1):
            for j in range(0, length_2 - 1):
                quadrilateral_polygons.append(volmdlr.wires.ClosedPolygon2D(
                    [self.lists_points[i][j],
                     self.lists_points[i + 1][j],
                     self.lists_points[i + 1][j + 1],
                     self.lists_points[i][j + 1]]))

        return quadrilateral_polygons

    @property
    def limits_xy(self):
        """
        find the limits (min, max) of points along x & y direction_axis
        """

        x_limits = (self.lists_points[0][0].x, self.lists_points[-1][-1].x)
        x_min, x_max = min(x_limits), max(x_limits)
        y_limits = (self.lists_points[0][0].y, self.lists_points[-1][-1].y)
        y_min, y_max = min(y_limits), max(y_limits)

        return ((x_min, x_max), (y_min, y_max))

    # @property
    # def x_limits(self):
    #     """
    #     find the limits (min, max) of points along x_direction_axis
    #     """

    #     x_limits = (self.lists_points[0][0].x, self.lists_points[-1][-1].x)
    #     x_min, x_max = min(x_limits), max(x_limits)

    #     return (x_min, x_max)

    # @property
    # def y_limits(self):
    #     """
    #     find the limits (min, max) of points along y_direction_axis
    #     """

    #     y_limits = (self.lists_points[0][0].y, self.lists_points[-1][-1].y)
    #     y_min, y_max = min(y_limits), max(y_limits)

    #     return (y_min, y_max)

    @property
    def points(self):
        """
        return all the points in lists_points in just one list

        Returns
        -------
        points: list[volmdlr.Point2D]

        """

        points = []
        for list_point in self.lists_points:
            points.extend(list_point)
        return points

    @property
    def points_xy(self):
        """
        find how many points there are along x & y direction_axis
        """

        index = self.find_direction_index(direction_axis='x')
        if index == 0:
            points_x = len(self.lists_points[0])
            points_y = len(self.lists_points)
        else:
            points_x = len(self.lists_points)
            points_y = len(self.lists_points[0])

        return (points_x, points_y)

    # @property
    # def points_x(self):
    #     """
    #     find how many points there are along x_direction_axis
    #     """

    #     index = self.find_direction_index(direction_axis = 'x')
    #     if index == 0:
    #         points_x = len(self.lists_points[0])
    #     else:
    #         points_x = len(self.lists_points)

    #     return points_x

    # @property
    # def points_y(self):
    #     """
    #     find how many points there are along y_direction_axis
    #     """

    #     index = self.find_direction_index(direction_axis = 'y')
    #     if index == 0:
    #         points_y = len(self.lists_points[0])
    #     else:
    #         points_y = len(self.lists_points)

    #     return points_y
